import numbers, warnings

from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import LabelEncoder, OneHotEncoder
from sklearn.metrics import log_loss
from sklearn.exceptions import ConvergenceWarning
from sklearn.utils import (check_X_y, compute_class_weight,
                           check_consistent_length, check_array)
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.extmath import safe_sparse_dot, log_logistic, squared_norm
from sklearn.preprocessing import StandardScaler

from scipy.optimize import minimize
from scipy import optimize
from scipy.special import expit, logsumexp

import numpy as np

from .base import AbstractUoIGeneralizedLinearRegressor
from ..utils import sigmoid, softmax
from ..lbfgs import fmin_lbfgs, AllZeroLBFGSError


class UoI_L1Logistic(AbstractUoIGeneralizedLinearRegressor, LogisticRegression):
    r"""UoI\ :sub:`L1-Logistic` model.

    Parameters
    ----------
    n_boots_sel : int
        The number of data bootstraps to use in the selection module.
        Increasing this number will make selection more strict.
    n_boots_est : int
        The number of data bootstraps to use in the estimation module.
        Increasing this number will relax selection and decrease variance.
    n_lambdas : int
        The number of regularization values to use for selection.
    alpha : list or ndarray
        The parameter that trades off L1 versus L2 regularization for a given
        lambda.
    selection_frac : float
        The fraction of the dataset to use for training in each resampled
        bootstrap, during the selection module. Small values of this parameter
        imply larger "perturbations" to the dataset.
    estimation_frac : float
        The fraction of the dataset to use for training in each resampled
        bootstrap, during the estimation module. The remaining data is used
        to obtain validation scores. Small values of this parameters imply
        larger "perturbations" to the dataset.
    estimation_target : string, "train" | "test"
        Decide whether to assess the estimation_score on the train
        or test data across each bootstrap. By deafult, a sensible
        choice is made based on the chosen estimation_score
    stability_selection : int, float, or array-like
        If int, treated as the number of bootstraps that a feature must
        appear in to guarantee placement in selection profile. If float,
        must be between 0 and 1, and is instead the proportion of
        bootstraps. If array-like, must consist of either ints or floats
        between 0 and 1. In this case, each entry in the array-like object
        will act as a separate threshold for placement in the selection
        profile.
    estimation_score : string, "acc" | "log" | "AIC", | "AICc" | "BIC"
        Objective used to choose the best estimates per bootstrap.
    multi_class : string, "auto" | "multinomial"
        For "multinomial" the loss minimised is the multinomial loss fit across
        the entire probability distribution, even when the data is binary.
        "auto" selects binary if the data is binary, and otherwise selects
        "multinomial".
    warm_start : bool
        When set to ``True``, reuse the solution of the previous call to fit as
        initialization, otherwise, just erase the previous solution
    eps : float
        Length of the L1 path. eps=1e-5 means that alpha_min / alpha_max = 1e-5
    fit_intercept : bool
        Whether to calculate the intercept for this model. If set to False, no
        intercept will be used in calculations (e.g. data is expected to be
        already centered).
    standardize : bool
        If True, the regressors X will be standardized before regression by
        subtracting the mean and dividing by their standard deviations.
    shared_support : bool
        For models with more than one output (multinomial logistic regression)
        this determines whether all outputs share the same support or can
        have independent supports.
    max_iter : int
        Maximum number of iterations for iterative fitting methods.
    tol : float
        Stopping criteria for solver.
    random_state : int, RandomState instance, or None
        The seed of the pseudo random number generator that selects a random
        feature to update.  If int, random_state is the seed used by the random
        number generator; If RandomState instance, random_state is the random
        number generator; If None, the random number generator is the
        RandomState instance used by `np.random`.
    comm : MPI communicator
        If passed, the selection and estimation steps are parallelized.

    Attributes
    ----------
    coef_ : array, shape (n_features,) or (n_targets, n_features)
        Estimated coefficients for the linear regression problem.
    intercept_ : float
        Independent term in the linear model.
    supports_ : array, shape
        boolean array indicating whether a given regressor (column) is selected
        for estimation for a given regularization parameter value (row).
    """

    metrics = AbstractUoIGeneralizedLinearRegressor._valid_estimation_metrics
    _valid_estimation_metrics = metrics + ('acc',)

    def __init__(self, n_boots_sel=24, n_boots_est=24, selection_frac=0.9,
                 estimation_frac=0.9, n_C=48, stability_selection=1.,
                 estimation_score='acc', estimation_target=None,
                 multi_class='auto', shared_support=True, warm_start=False,
                 eps=1e-5, fit_intercept=True, standardize=True,
                 max_iter=10000, tol=1e-3, random_state=None, comm=None,
                 logger=None):
        super(UoI_L1Logistic, self).__init__(
            n_boots_sel=n_boots_sel,
            n_boots_est=n_boots_est,
            selection_frac=selection_frac,
            estimation_frac=estimation_frac,
            stability_selection=stability_selection,
            estimation_score=estimation_score,
            estimation_target=estimation_target,
            random_state=random_state,
            fit_intercept=fit_intercept,
            standardize=standardize,
            shared_support=shared_support,
            comm=comm,
            logger=logger)
        self.n_C = n_C
        self.Cs = None
        self.multi_class = multi_class
        self.eps = eps
        self.tol = tol
        self.solver = 'lbfgs'
        self._selection_lm = MaskedCoefLogisticRegression(
            penalty='l1',
            max_iter=max_iter,
            warm_start=warm_start,
            multi_class=multi_class,
            fit_intercept=fit_intercept,
            tol=tol)

        self._estimation_lm = MaskedCoefLogisticRegression(
            C=np.inf,
            multi_class=multi_class,
            fit_intercept=fit_intercept,
            max_iter=max_iter,
            tol=tol)

    def get_reg_params(self, X, y):
        input_dim = X.shape[1]
        if self.Cs is None:
            if self.output_dim == 1:
                w = np.zeros(input_dim)
                if self.fit_intercept:
                    intercept = LogisticInterceptFitterNoFeatures(
                        y, 1).intercept_
                    w = np.concatenate([w, np.atleast_1d(intercept)])
                _, grad = _logistic_loss_and_grad(w, X, y, 0., None)
            else:
                w = np.zeros((self.output_dim, input_dim))
                yp = OneHotEncoder(categories=[range(self.output_dim)],
                                   sparse=False).fit_transform(y[:, np.newaxis])
                if self.fit_intercept:
                    intercept = LogisticInterceptFitterNoFeatures(
                        y, self.output_dim).intercept_
                    w = np.concatenate([w, intercept[:, np.newaxis]], axis=1)
                _, grad, _ = _multinomial_loss_grad(w, X, yp, 0., None,
                                                    np.ones(X.shape[0]))
            alpha_max = abs(grad[:-1]).max()
            logC = -np.log10(alpha_max)
            self.Cs = np.logspace(logC, logC - np.log10(self.eps), self.n_C)
        ret = list()
        for c in self.Cs:
            ret.append(dict(C=c))
        return ret

    def _fit_intercept_no_features(self, y):
        """"Fit a model with only an intercept.

        This is used in cases where the model has no support selected.
        """
        return LogisticInterceptFitterNoFeatures(y, self.output_dim)

    def _fit_intercept(self, X, y):
        """"Fit a model with an intercept and fixed coefficients.

        This is used to re-fit the intercept after the coefficients are
        estimated.
        """
        if self.fit_intercept:
            self.intercept_ = fit_intercept_fixed_coef(X, self.coef_, y,
                                                       self.output_dim)
        else:
            self.intercept_ = np.zeros(self.output_dim)

    def _pre_fit(self, X, y):
        X, y = super()._pre_fit(X, y)
        le = LabelEncoder()
        y = le.fit_transform(y)
        self.classes_ = le.classes_
        if self.classes_.size > 2:
            self.output_dim = self.classes_.size
        elif self.multi_class == 'multinomial':
            self.output_dim = 2
        else:
            self.output_dim = 1
        return X, y


def fit_intercept_fixed_coef(X, coef_, y, output_dim):
    """Optimize the likelihood w.r.t. the intercept for a logistic model."""
    if output_dim == 1:
        def f_df(intercept):
            py = sigmoid(X.dot(coef_.T) + intercept)
            dfdb = py.mean() - y.mean()
            return log_loss(y, py), np.atleast_1d(dfdb)

        opt = minimize(f_df, np.atleast_1d(np.zeros(1)),
                       method='L-BFGS-B', jac=True)
        return opt.x
    else:
        def f_df(short_intercept):
            intercept = np.concatenate([np.atleast_1d(1.), short_intercept])
            py = softmax(X.dot(coef_.T) + intercept)

            def dlogpi_dintk(ii, pyi):
                if ii == 0:
                    return -pyi[1:]
                else:
                    rval = np.eye(output_dim - 1)[ii - 1]
                    rval -= pyi[1:]
                    return rval

            dfdb = np.zeros_like(short_intercept)
            for yi, pyi in zip(y, py):
                dfdb -= dlogpi_dintk(yi, pyi) / y.size
            return log_loss(y, py, labels=np.arange(output_dim)), dfdb
        opt = minimize(f_df, np.atleast_1d(np.zeros(output_dim - 1)),
                       method='L-BFGS-B', jac=True)
        intercept = np.concatenate([np.atleast_1d(1.), opt.x])
        return intercept - intercept.max()


class LogisticInterceptFitterNoFeatures(object):
    """Intercept-only bernoulli logistic regression.

    Parameters
    ----------
    y : ndarray
        Class labels.
    """
    def __init__(self, y, output_dim):
        self.output_dim = output_dim
        eps = 1e-10
        if output_dim == 1:
            p = y.mean(axis=0)
            p = np.minimum(np.maximum(p, eps), 1 - eps)
            self.intercept_ = np.log(p / (1. - p))
        else:
            py = np.equal(y[:, np.newaxis],
                          np.arange(self.output_dim)[np.newaxis]).mean(axis=0)
            n_included = np.count_nonzero(py)
            if n_included < self.output_dim:
                new_mass = eps * (self.output_dim - n_included)
                py *= (1. - new_mass)
                py[np.equal(py, 0.)] = eps
            intercept = np.log(py)
            self.intercept_ = intercept - intercept.max()

    def predict(self, X, mask=None):
        n_samples = X.shape[0]
        if self.output_dim == 1:
            return np.tile(int(self.intercept_ >= 0.), n_samples)
        else:
            return np.tile(int(np.argmax(self.intercept_)), n_samples)

    def predict_proba(self, X, mask=None):
        n_samples = X.shape[0]
        if self.output_dim == 1:
            return np.tile(sigmoid(self.intercept_), n_samples)
        else:
            return np.tile(softmax(self.intercept_)[np.newaxis], (n_samples, 1))


class MaskedCoefLogisticRegression(LogisticRegression):
    """Logistic regression with a binary mask on the coef.

    Parameters
    ----------
    penalty : str
        Type of regularization: 'l1' or 'l2'.
    tol : float, optional (default=1e-4)
        Tolerance for stopping criteria.
    C : float, optional (default=1.0)
        Inverse of regularization strength; must be a positive float.
        Like in support vector machines, smaller values specify stronger
        regularization.
    fit_intercept : bool, optional (default=True)
        Specifies if a constant (a.k.a. bias or intercept) should be
        added to the decision function.
    standardize : bool, default False
        If True, centers the design matrix across samples and rescales them to
        have standard deviation of 1.
    class_weight : dict or 'balanced', optional (default=None)
        Weights associated with classes in the form ``{class_label: weight}``.
        If not given, all classes are supposed to have weight one.
        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``.
        Note that these weights will be multiplied with sample_weight (passed
        through the fit method) if sample_weight is specified.
    max_iter : int, optional (default=100)
        Maximum number of iterations taken for the solvers to converge.
    multi_class : str, {'multinomial', 'auto'}, optional (default='auto')
        For 'multinomial' the loss minimised is the multinomial loss fit
        across the entire probability distribution, *even when the data is
        binary*. 'auto' selects binary if the data is binary,
        and otherwise selects 'multinomial'.
    verbose : int, optional (default=0)
        For the liblinear and lbfgs solvers set verbose to any positive
        number for verbosity.
    warm_start : bool, optional (default=False)
        When set to True, reuse the solution of the previous call to fit as
        initialization, otherwise, just erase the previous solution.
        Useless for liblinear solver.
    """
    def __init__(self, penalty='l2', tol=1e-3, C=1.,
                 fit_intercept=True, standardize=False, class_weight=None,
                 max_iter=10000,
                 multi_class='auto', verbose=0, warm_start=False):
        if multi_class not in ('multinomial', 'auto'):
            raise ValueError("multi_class should be 'multinomial' or " +
                             "'auto'. Got %s." % multi_class)
        super().__init__(penalty=penalty, tol=tol, C=C,
                         fit_intercept=fit_intercept,
                         class_weight=class_weight,
                         max_iter=max_iter,
                         multi_class=multi_class, verbose=verbose,
                         warm_start=warm_start)
        self.standardize = standardize

    def fit(self, X, y, sample_weight=None, coef_mask=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.
        y : array-like, shape (n_samples,)
            Target vector relative to X.
        sample_weight : array-like, shape (n_samples,) optional
            Array of weights that are assigned to individual samples.
            If not provided, then each sample is given unit weight.
        coef_mask : array-like, shape (n_features), (n_classes, n_features)
                    optional
            Masking array for coef.

        Returns
        -------
        self : object
        """
        if not isinstance(self.C, numbers.Number) or self.C < 0:
            raise ValueError("Penalty term must be positive; got (C=%r)"
                             % self.C)
        if not isinstance(self.max_iter, numbers.Number) or self.max_iter < 0:
            raise ValueError("Maximum number of iteration must be positive;"
                             " got (max_iter=%r)" % self.max_iter)
        if not isinstance(self.tol, numbers.Number) or self.tol < 0:
            raise ValueError("Tolerance for stopping criteria must be "
                             "positive; got (tol=%r)" % self.tol)

        _dtype = np.float64

        X, y = self._pre_fit(X, y)

        X, y = check_X_y(X, y, accept_sparse='csr', dtype=_dtype, order="C",
                         accept_large_sparse=True)
        check_classification_targets(y)
        self.classes_ = np.unique(y)
        n_samples, n_features = X.shape

        multi_class = self.multi_class
        if multi_class == 'auto':
            if len(self.classes_) > 2:
                multi_class = 'multinomial'
            else:
                multi_class = 'ovr'

        n_classes = len(self.classes_)
        classes_ = self.classes_
        if n_classes < 2:
            raise ValueError("This solver needs samples of at least 2 classes"
                             " in the data, but the data contains only one"
                             " class: %r" % classes_[0])
        if len(self.classes_) == 2 and multi_class == 'ovr':
            n_classes = 1
            classes_ = classes_[1:]
        if multi_class == 'multinomial' and coef_mask is not None:
            coef_mask = coef_mask.reshape(n_classes, -1)

        if self.warm_start:
            warm_start_coef = getattr(self, 'coef_', None)
        else:
            warm_start_coef = None
        if warm_start_coef is not None and self.fit_intercept:
            warm_start_coef = np.append(warm_start_coef,
                                        self.intercept_[:, np.newaxis],
                                        axis=1)

        self.intercept_ = np.zeros(n_classes)

        fold_coefs_ = _logistic_regression_path(
            X, y, Cs=[self.C],
            fit_intercept=self.fit_intercept,
            tol=self.tol, verbose=self.verbose,
            multi_class=multi_class, max_iter=self.max_iter,
            class_weight=self.class_weight, penalty=self.penalty,
            check_input=False,
            coef=warm_start_coef,
            sample_weight=sample_weight, coef_mask=coef_mask)

        fold_coefs_, _, self.n_iter_ = fold_coefs_
        self.coef_ = np.asarray(fold_coefs_)
        self.coef_ = self.coef_.reshape(n_classes, n_features +
                                        int(self.fit_intercept))

        if self.fit_intercept:
            self.intercept_ = self.coef_[:, -1]
            self.coef_ = self.coef_[:, :-1]

        self._post_fit(X, y)

        return self

    def _pre_fit(self, X, y):
        if self.standardize:
            self._X_scaler = StandardScaler()
            X = self._X_scaler.fit_transform(X)
        le = LabelEncoder()
        y = le.fit_transform(y)
        self.classes_ = le.classes_
        if self.classes_.size > 2:
            self.output_dim = self.classes_.size
        elif self.multi_class == 'multinomial':
            self.output_dim = 2
        else:
            self.output_dim = 1
        return X, y

    def _post_fit(self, X, y):
        """Perform class-specific cleanup for fit().
        """
        if self.standardize:
            sX = self._X_scaler
            self.coef_ /= sX.scale_[np.newaxis]


def _logistic_regression_path(X, y, Cs=48, fit_intercept=True,
                              max_iter=100, tol=1e-4, verbose=0, coef=None,
                              class_weight=None, penalty='l2',
                              multi_class='auto',
                              check_input=True,
                              sample_weight=None,
                              l1_ratio=None, coef_mask=None):
    """Compute a Logistic Regression model for a list of regularization
    parameters.

    Parameters
    ----------
    X : array-like or sparse matrix, shape (n_samples, n_features)
        Input data.
    y : array-like, shape (n_samples,) or (n_samples, n_targets)
        Input data, target values.
    Cs : int | array-like, shape (n_cs,)
        List of values for the regularization parameter or integer specifying
        the number of regularization parameters that should be used. In this
        case, the parameters will be chosen in a logarithmic scale between
        1e-4 and 1e4.
    fit_intercept : bool
        Whether to fit an intercept for the model. In this case the shape of
        the returned array is (n_cs, n_features + 1).
    max_iter : int
        Maximum number of iterations for the solver.
    tol : float
        Stopping criterion. For the newton-cg and lbfgs solvers, the iteration
        will stop when ``max{|g_i | i = 1, ..., n} <= tol``
        where ``g_i`` is the i-th component of the gradient.
    verbose : int
        For the liblinear and lbfgs solvers set verbose to any positive
        number for verbosity.
    coef : array-like, shape (n_features,), default None
        Initialization value for coefficients of logistic regression.
        Useless for liblinear solver.
    class_weight : dict or 'balanced', optional
        Weights associated with classes in the form ``{class_label: weight}``.
        If not given, all classes are supposed to have weight one.
        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``.
        Note that these weights will be multiplied with sample_weight (passed
        through the fit method) if sample_weight is specified.
    multi_class : str, {'multinomial', 'auto'}, default: 'auto'
        For 'multinomial' the loss minimised is the multinomial loss fit
        across the entire probability distribution, *even when the data is
        binary*. 'auto' selects binary if the data is binary
        and otherwise selects 'multinomial'.
    check_input : bool, default True
        If False, the input arrays X and y will not be checked.
    sample_weight : array-like, shape(n_samples,) optional
        Array of weights that are assigned to individual samples.
        If not provided, then each sample is given unit weight.
    coef_mask : array-like, shape (n_features), (n_classes, n_features) optional
        Masking array for coef.
    Returns
    -------
    coefs : ndarray, shape (n_cs, n_features) or (n_cs, n_features + 1)
        List of coefficients for the Logistic Regression model. If
        fit_intercept is set to True then the second dimension will be
        n_features + 1, where the last item represents the intercept. For
        ``multiclass='multinomial'``, the shape is (n_classes, n_cs,
        n_features) or (n_classes, n_cs, n_features + 1).
    Cs : ndarray
        Grid of Cs used for cross-validation.
    n_iter : array, shape (n_cs,)
        Actual number of iteration for each Cs.
    """
    if isinstance(Cs, numbers.Integral):
        Cs = np.logspace(-4, 4, Cs)

    # Preprocessing.
    if check_input:
        X = check_array(X, accept_sparse='csr', dtype=np.float64,
                        accept_large_sparse=True)
        y = check_array(y, ensure_2d=False, dtype=None)
        check_consistent_length(X, y)
    _, n_features = X.shape

    classes = np.unique(y)
    n_classes = len(classes)

    if multi_class == 'auto':
        if n_classes > 2:
            multi_class = 'multinomial'
        else:
            multi_class = 'ovr'

    # If sample weights exist, convert them to array (support for lists)
    # and check length
    # Otherwise set them to 1 for all examples
    if sample_weight is not None:
        sample_weight = np.array(sample_weight, dtype=X.dtype, order='C')
        check_consistent_length(y, sample_weight)
    else:
        sample_weight = np.ones(X.shape[0], dtype=X.dtype)

    # If class_weights is a dict (provided by the user), the weights
    # are assigned to the original labels. If it is "balanced", then
    # the class_weights are assigned after masking the labels with a OvR.
    le = LabelEncoder()
    if isinstance(class_weight, dict) or multi_class == 'multinomial':
        class_weight_ = compute_class_weight(class_weight, classes=classes, y=y)
        sample_weight *= class_weight_[le.fit_transform(y)]

    # For doing a ovr, we need to mask the labels first. for the
    # multinomial case this is not necessary.
    if multi_class == 'ovr':
        coef_size = n_features
        w0 = np.zeros(n_features + int(fit_intercept), dtype=X.dtype)
        mask_classes = np.array([-1, 1])
        mask = (y == 1)
        y_bin = np.ones(y.shape, dtype=X.dtype)
        y_bin[~mask] = -1.
        # for compute_class_weight

        if class_weight == "balanced":
            class_weight_ = compute_class_weight(class_weight,
                                                 classes=mask_classes,
                                                 y=y_bin)
            sample_weight *= class_weight_[le.fit_transform(y_bin)]

    else:
        coef_size = classes.size * n_features
        lbin = OneHotEncoder(categories=[range(classes.size)], sparse=False)
        Y_multi = lbin.fit_transform(y[:, np.newaxis])
        if Y_multi.shape[1] == 1:
            Y_multi = np.hstack([1 - Y_multi, Y_multi])
        w0 = np.zeros((classes.size, n_features + int(fit_intercept)),
                      dtype=X.dtype)
        w0[:, -1] = LogisticInterceptFitterNoFeatures(y,
                                                      classes.size).intercept_

    if coef is not None:
        # it must work both giving the bias term and not
        if multi_class == 'ovr':
            if coef.size not in (n_features, w0.size):
                raise ValueError(
                    'Initialization coef is of shape %d, expected shape '
                    '%d or %d' % (coef.size, n_features, w0.size))
            w0[:coef.size] = coef
        else:
            w0[:, :coef.shape[1]] = coef

    # Mask initial array
    if coef_mask is not None:
        if multi_class == 'ovr':
            w0[:n_features] *= coef_mask
        else:
            w0[:, :n_features] *= coef_mask

    if multi_class == 'multinomial':
        # fmin_l_bfgs_b and newton-cg accepts only ravelled parameters.
        target = Y_multi
        if penalty == 'l2':
            w0 = w0.ravel()
            if coef_mask is not None:
                x0 = np.zeros_like(w0)

            def func(x, *args):
                mask = args[3]
                if mask is not None:
                    x0[mask] = x
                    args = args[:3] + (None,) + (args[-1],)
                    f, df = _multinomial_loss_grad(x0, *args)[0:2]
                    df = df[mask]
                else:
                    f, df = _multinomial_loss_grad(x, *args)[0:2]
                return f, df
        else:
            w0 = w0.T.ravel().copy()

            def inner_func(x, *args):
                return _multinomial_loss_grad(x, *args)[0:2]

            def func(x, g, *args):
                x = x.reshape(-1, classes.size).T.ravel()
                loss, grad = inner_func(x, *args)
                grad = grad.reshape(classes.size, -1).T.ravel()
                g[:] = grad
                return loss
    else:
        target = y_bin
        if penalty == 'l2':
            x0 = np.zeros_like(w0)

            def func(x, *args):
                mask = args[3]
                if mask is not None:
                    x0[mask] = x
                    args = args[:3] + (None,) + (args[-1],)
                    f, df = _logistic_loss_and_grad(x0, *args)
                    df = df[mask]
                else:
                    f, df = _logistic_loss_and_grad(x, *args)
                return f, df
        else:
            def func(x, g, *args):
                loss, grad = _logistic_loss_and_grad(x, *args)
                g[:] = grad
                return loss

    coefs = list()
    n_iter = np.zeros(len(Cs), dtype=np.int32)
    for i, C in enumerate(Cs):
        if coef_mask is None or coef_mask.sum():
            iprint = [-1, 50, 1, 100, 101][
                np.searchsorted(np.array([0, 1, 2, 3]), verbose)]
            if penalty == 'l2':
                if coef_mask is None:
                    w0, loss, info = optimize.fmin_l_bfgs_b(
                        func, w0, fprime=None,
                        args=(X, target, 1. / C, coef_mask, sample_weight),
                        iprint=iprint, pgtol=tol, maxiter=max_iter)
                else:
                    if fit_intercept:
                        if multi_class == 'multinomial':
                            mask = [coef_mask,
                                    np.ones(n_classes)[:, np.newaxis]]
                            mask = np.concatenate(mask, axis=1)
                        else:
                            mask = np.concatenate([coef_mask, np.ones(1)])
                    else:
                        mask = coef_mask
                    mask = np.nonzero(mask.ravel())[0]
                    wp = w0[mask]
                    wp, loss, info = optimize.fmin_l_bfgs_b(
                        func, wp, fprime=None,
                        args=(X, target, 1. / C, mask, sample_weight),
                        iprint=iprint, pgtol=tol, maxiter=max_iter)
                    w0 = np.zeros_like(w0)
                    w0[mask] = wp

            else:
                zeros_seen = [0]

                def zero_coef(x, *args):
                    if multi_class == 'multinomial':
                        x = x.reshape(-1, classes.size)[:-1]
                    else:
                        x = x[:-1]
                    now_zeros = np.array_equiv(x, 0.)
                    if now_zeros:
                        zeros_seen[0] += 1
                    else:
                        zeros_seen[0] = 0
                    if zeros_seen[0] > 1:
                        return -2048
                try:
                    args = (X, target, 0., coef_mask, sample_weight)
                    w0 = fmin_lbfgs(func, w0, orthantwise_c=1. / C,
                                    args=args,
                                    max_iterations=max_iter,
                                    epsilon=tol,
                                    orthantwise_end=coef_size,
                                    progress=zero_coef)
                except AllZeroLBFGSError:
                    w0 *= 0.
                info = None
            if info is not None and info["warnflag"] == 1:
                warnings.warn("lbfgs failed to converge. Increase the number "
                              "of iterations.", ConvergenceWarning)
            # In scipy <= 1.0.0, nit may exceed maxiter.
            # See https://github.com/scipy/scipy/issues/7854.
            if info is None:
                n_iter_i = -1
            else:
                n_iter_i = min(info['nit'], max_iter)

            n_iter[i] = n_iter_i

        if multi_class == 'multinomial':
            n_classes = max(2, classes.size)
            if penalty == 'l2':
                w0 = np.reshape(w0, (n_classes, -1))
            else:
                w0 = np.reshape(w0, (-1, n_classes)).T
        coefs.append(w0.copy())

    return np.array(coefs), np.array(Cs), n_iter


def _multinomial_loss(w, X, Y, alpha, sample_weight):
    """Computes multinomial loss and class probabilities.

    Parameters
    ----------
    w : ndarray, shape (n_classes * n_features,) or
        (n_classes * (n_features + 1),)
        Coefficient vector.
    X : {array-like, sparse matrix}, shape (n_samples, n_features)
        Training data.
    Y : ndarray, shape (n_samples, n_classes)
        Transformed labels according to the output of LabelBinarizer.
    alpha : float
        Regularization parameter. alpha is equal to 1 / C.
    sample_weight : array-like, shape (n_samples,)
        Array of weights that are assigned to individual samples.

    Returns
    -------
    loss : float
        Multinomial loss.
    p : ndarray, shape (n_samples, n_classes)
        Estimated class probabilities.
    w : ndarray, shape (n_classes, n_features)
        Reshaped param vector excluding intercept terms.
    """
    n_classes = Y.shape[1]
    n_samples, n_features = X.shape
    fit_intercept = w.size == (n_classes * (n_features + 1))
    w = w.reshape(n_classes, -1)
    sample_weight = sample_weight[:, np.newaxis]
    if fit_intercept:
        intercept = w[:, -1]
        w = w[:, :-1]
    else:
        intercept = 0
    p = safe_sparse_dot(X, w.T)
    p += intercept
    p -= logsumexp(p, axis=1)[:, np.newaxis]
    loss = -(sample_weight * Y * p).sum() / n_samples
    loss += 0.5 * alpha * squared_norm(w)
    p = np.exp(p, p)
    return loss, p, w


def _logistic_loss_and_grad(w, X, y, alpha, mask, sample_weight=None):
    """Computes the logistic loss and gradient.

    Parameters
    ----------
    w : ndarray, shape (n_features,) or (n_features + 1,)
        Coefficient vector.
    X : {array-like, sparse matrix}, shape (n_samples, n_features)
        Training data.
    y : ndarray, shape (n_samples,)
        Array of labels.
    alpha : float
        Regularization parameter. alpha is equal to 1 / C.
    mask : array-like, shape (n_features), (n_classes, n_features) optional
        Masking array for coef.
    sample_weight : array-like, shape (n_samples,) optional
        Array of weights that are assigned to individual samples.
        If not provided, then each sample is given unit weight.

    Returns
    -------
    out : float
        Logistic loss.
    grad : ndarray, shape (n_features,) or (n_features + 1,)
        Logistic gradient.
    """
    n_samples, n_features = X.shape
    if mask is not None:
        w[:n_features] *= mask
    grad = np.empty_like(w)

    w, c, yz = _intercept_dot(w, X, y)

    if sample_weight is None:
        sample_weight = np.ones(n_samples)

    # Logistic loss is the negative of the log of the logistic function.
    out = -np.sum(sample_weight * log_logistic(yz)) / n_samples
    out += .5 * alpha * np.dot(w, w)

    z = expit(yz)
    z0 = sample_weight * (z - 1) * y

    grad[:n_features] = (safe_sparse_dot(X.T, z0) / n_samples) + alpha * w
    if mask is not None:
        grad[:n_features] *= mask

    # Case where we fit the intercept.
    if grad.shape[0] > n_features:
        grad[-1] = z0.sum() / n_samples
    return out, grad


def _multinomial_loss_grad(w, X, Y, alpha, mask, sample_weight):
    """Computes the multinomial loss, gradient and class probabilities.

    Parameters
    ----------
    w : ndarray, shape (n_classes * n_features,) or
        (n_classes * (n_features + 1),)
        Coefficient vector.
    X : {array-like, sparse matrix}, shape (n_samples, n_features)
        Training data.
    Y : ndarray, shape (n_samples, n_classes)
        Transformed labels according to the output of LabelBinarizer.
    alpha : float
        Regularization parameter. alpha is equal to 1 / C.
    mask : array-like, shape (n_features), (n_classes, n_features) optional
        Masking array for coef.
    sample_weight : array-like, shape (n_samples,)
        Array of weights that are assigned to individual samples.

    Returns
    -------
    loss : float
        Multinomial loss.
    grad : ndarray, shape (n_classes * n_features,) or
        (n_classes * (n_features + 1),)
        Ravelled gradient of the multinomial loss.
    p : ndarray, shape (n_samples, n_classes)
        Estimated class probabilities
    """
    n_classes = Y.shape[1]
    n_samples, n_features = X.shape
    fit_intercept = (w.size == n_classes * (n_features + 1))
    if mask is not None:
        w = w.reshape(n_classes, n_features + bool(fit_intercept))
        w[:, :n_features] *= mask
        w = w.ravel()
    grad = np.zeros((n_classes, n_features + bool(fit_intercept)),
                    dtype=X.dtype)
    loss, p, w = _multinomial_loss(w, X, Y, alpha, sample_weight)
    sample_weight = sample_weight[:, np.newaxis]
    diff = sample_weight * (p - Y)
    grad[:, :n_features] = safe_sparse_dot(diff.T, X) / n_samples
    grad[:, :n_features] += alpha * w
    if mask is not None:
        grad[:, :n_features] *= mask
    if fit_intercept:
        grad[:, -1] = diff.sum(axis=0) / n_samples
    return loss, grad.ravel(), p


def _intercept_dot(w, X, y):
    """Computes y * np.dot(X, w).
    It takes into consideration if the intercept should be fit or not.
    Parameters
    ----------
    w : ndarray of shape (n_features,) or (n_features + 1,)
        Coefficient vector.
    X : {array-like, sparse matrix} of shape (n_samples, n_features)
        Training data.
    y : ndarray of shape (n_samples,)
        Array of labels.
    Returns
    -------
    w : ndarray of shape (n_features,)
        Coefficient vector without the intercept weight (w[-1]) if the
        intercept should be fit. Unchanged otherwise.
    c : float
        The intercept.
    yz : float
        y * np.dot(X, w).
    """
    c = 0.
    if w.size == X.shape[1] + 1:
        c = w[-1]
        w = w[:-1]

    z = safe_sparse_dot(X, w) + c
    yz = y * z
    return w, c, yz



import numpy as np

from .base import AbstractUoIGeneralizedLinearRegressor

from pyuoi import utils

from sklearn.base import BaseEstimator
from sklearn.preprocessing import StandardScaler
from sklearn.utils.validation import check_is_fitted

from ..lbfgs import fmin_lbfgs


class Poisson(BaseEstimator):
    r"""Generalized Linear Model with exponential link function (i.e. Poisson)
    trained with L1/L2 regularizer (i.e. Elastic net penalty).

    The log-likelihood of the Poisson GLM is optimized by performing coordinate
    descent on a linearized quadratic approximation. See Chapter 5 of Hastie,
    Tibshirani, and Wainwright (2016) for more details.

    Parameters
    ----------
    alpha : float
        Constant that multiplies the L1 term. Defaults to 1.0.
    l1_ratio : float
        Float between 0 and 1 acting as a scaling between l1 and l2 penalties.
        For ``l1_ratio = 0`` the penalty is an L2 penalty. For ``l1_ratio = 1``
        it is an L1 penalty. For ``0 < l1_ratio < 1``, the penalty is a
        combination of L1 and L2 penalties.
    fit_intercept : bool
        Whether to fit an intercept or not.
    standardize : bool
        If ``True``, centers the design matrix across samples and rescales them
        to have standard deviation of 1.
    tol : float
        The tolerance for the optimization: if the updates are smaller than
        ``tol``, the optimization code checks the dual gap for optimality and
        continues until it is smaller than ``tol``.
    warm_start : bool
        When set to ``True``, reuse the solution of the previous call to
        fit as initialization, otherwise, just erase the previous solution.
    solver : string, 'lbfgs' | 'cd'
        The solver to use. Options are 'lbfgs' (orthant-wise LBFGS) and 'cd'
        (coordinate descent).

    Attributes
    ----------
    coef_ : ndarray, shape (n_features,)
        The fitted parameter vector.
    intercept_ : float
        The fitted intercept.
    """
    def __init__(self, alpha=1.0, l1_ratio=1., fit_intercept=True,
                 standardize=False, max_iter=1000, tol=1e-5, warm_start=False,
                 solver='lbfgs'):
        self.alpha = alpha
        self.l1_ratio = l1_ratio
        self.fit_intercept = fit_intercept
        self.standardize = standardize
        self.max_iter = max_iter
        self.tol = tol
        self.warm_start = warm_start
        self.solver = solver

    def fit(self, X, y, sample_weight=None):
        """Fit the Poisson GLM.

        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            The design matrix.
        y : ndarray, shape (n_samples,)
            Response vector. Will be cast to X's dtype if necessary. Currently,
            this implementation does not handle multiple response variables.
        sample_weight : array-like, shape (n_samples,)
            Array of weights assigned to the individual samples. If ``None``,
            then each sample is provided an equal weight.
        """
        self.n_samples, self.n_features = X.shape
        X, y = self._pre_fit(X, y)

        # check if the response variable is constant
        if np.unique(y).size == 1:
            self.coef_ = np.zeros(self.n_features)
            if np.any(y):
                self.intercept_ = np.log(y.mean())
            else:
                self.intercept_ = -np.inf
            return self

        if self.solver == 'lbfgs':
            coef = np.zeros(self.n_features)
            intercept = 0
            if self.warm_start:
                if hasattr(self, 'coef_'):
                    if self.standardize:
                        coef = self.coef_ * self._X_scaler.scale_
                    else:
                        coef = self.coef_
                if hasattr(self, 'intercept_') and self.fit_intercept:
                    intercept = self.intercept_

            if self.fit_intercept:
                coef = np.append(coef, intercept)

            # create lbfgs function
            def func(x, g, *args):
                loss, grad = _poisson_loss_and_grad(x, *args)
                g[:] = grad
                return loss

            l1_penalty = self.alpha * self.l1_ratio
            l2_penalty = self.alpha * (1 - self.l1_ratio)

            # orthant-wise lbfgs optimization
            coef = fmin_lbfgs(func, coef,
                              orthantwise_c=l1_penalty,
                              args=(X, y, l2_penalty, sample_weight),
                              max_iterations=self.max_iter,
                              epsilon=self.tol,
                              orthantwise_end=self.n_features)

            if self.fit_intercept:
                self.coef_ = coef[:self.n_features]
                self.intercept_ = coef[-1]
            else:
                self.coef_ = coef

        # coordinate descent
        elif self.solver == 'cd':
            self.coef_, intercept = self._cd(X=X, y=y,
                                             sample_weight=sample_weight)

            if self.fit_intercept:
                self.intercept_ = intercept

        else:
            raise ValueError('Solver not available.')

        self._post_fit(X, y)
        return self

    def predict(self, X):
        """Predicts the response variable given a design matrix. The output is
        the mode of the Poisson distribution.

        Parameters
        ----------
        X : array_like, shape (n_samples, n_features)
            Design matrix to predict on.

        Returns
        -------
        mode : array_like, shape (n_samples)
            The predicted response values, i.e. the modes.
        """
        check_is_fitted(self, ['coef_', 'intercept_'])
        mu = np.exp(self.intercept_ + np.dot(X, self.coef_))
        mode = np.floor(mu)
        return mode

    def predict_mean(self, X):
        """Calculates the mean response variable given a design matrix.

        Parameters
        ----------
        X : array_like, shape (n_samples, n_features)
            Design matrix to predict on.

        Returns
        -------
        mu : array_like, shape (n_samples)
            The predicted response values, i.e. the conditional means.
        """
        check_is_fitted(self, ['coef_', 'intercept_'])
        mu = np.exp(self.intercept_ + np.dot(X, self.coef_))
        return mu

    def _cd(self, X, y, sample_weight=None):
        """Performs coordinate descent on a dataset.

        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            The design matrix.
        y : ndarray, shape (n_samples,)
            The response vector.

        sample_weight : array-like, shape (n_samples,), default None
            Array of weights assigned to the individual samples. If None, then
            each sample is provided an equal weight.

        Returns
        -------
        coef : ndarray, shape (n_features,)
            The fitted coefficients.
        intercept : float
            The fitted intercept. If fit_intercept = False, this will be equal
            to zero.
        """
        n_samples, n_features = X.shape

        if sample_weight is None:
            sample_weight = np.ones(n_samples)

        # initialization of coef and intercepts
        coef = np.zeros(n_features)
        intercept = 0
        # warm start coefficients, if necessary
        if self.warm_start:
            if hasattr(self, 'coef_'):
                if self.standardize:
                    coef = self.coef_ * self._X_scaler.scale_
                else:
                    coef = self.coef_

            if self.fit_intercept:
                if hasattr(self, 'intercept_'):
                    intercept = self.intercept_

        # every coefficient is active unless there's a warm start
        if self.warm_start:
            active_idx = np.argwhere(coef != 0)
        else:
            active_idx = np.arange(self.n_features)

        prev_coef = np.copy(coef)
        # perform coordinate descent updates
        for iteration in range(self.max_iter):
            # linearize the log-likelihood
            w, z = self.adjusted_response(X, y, coef, intercept)
            # rescale weights by sample_weight
            w *= sample_weight

            # perform an update of coordinate descent
            coef, intercept = self._cd_sweep(
                coef=coef, intercept=intercept, X=X, w=w, z=z,
                active_idx=active_idx)

            # check convergence
            if np.max(np.abs(prev_coef - coef)) < self.tol:
                break

            prev_coef = np.copy(coef)

            # update the active features
            active_idx = np.argwhere(coef != 0).ravel()

        return coef, intercept

    def _cd_sweep(self, coef, X, w, z, active_idx, intercept=0):
        """Performs one sweep of coordinate descent updates over a set of
        'active' features.

        Parameters
        ----------
        coef : ndarray, shape (n_features,)
            The current estimates of the parameters.
        X : ndarray, shape (n_samples, n_features)
            The design matrix.
        w : ndarray, shape (n_samples,)
            The weights applied to each sample after linearization of the
            log-likelihood.
        z : ndarray, shape (n_samples,)
            The working response after linearization of the log-likelihood.
        active_idx : ndarray, shape (n_features,)
            An array of ints denoting the indices of features that should be
            updated.
        intercept : float
            The current estimate of the intercept.

        Returns
        -------
        coef : ndarray, shape (n_features,)
            The updated parameters after one coordinate descent sweep.
        intercept : float
            The update intercept after one coordinate descent sweep.
        """
        n_features = coef.size
        n_samples = X.shape[0]

        # intercept is not penalized
        if self.fit_intercept:
            z_hat = np.dot(X, coef)
            residuals = z - z_hat
            # update intercept
            intercept = np.dot(w, residuals) / np.sum(w)
        else:
            intercept = 0

        # iterate over the active features
        for idx in active_idx:
            # remove the current feature from the update rules
            mask = np.ones(n_features)
            mask[idx] = 0

            z_hat = intercept + np.dot(X, coef * mask)
            x = X[:, idx]
            # equation 5.44; Hastie, Tibshirani, Wainwright (2016)
            num_update = np.dot(w, x * (z - z_hat)) / n_samples
            den_update = np.dot(w, x**2) / n_samples
            # replace coefficients sequentially
            coef[idx] = \
                self.soft_threshold(num_update, self.l1_ratio * self.alpha) \
                / (den_update + self.alpha * (1 - self.l1_ratio))

        return coef, intercept

    def _pre_fit(self, X, y):
        """Perform standardization, if needed, before fitting."""
        if self.standardize:
            self._X_scaler = StandardScaler()
            X = self._X_scaler.fit_transform(X)
        return X, y

    def _post_fit(self, X, y):
        """Rescale coefficients, if needed, after fitting."""
        if self.standardize:
            sX = self._X_scaler
            if self.fit_intercept:
                self.intercept_ -= np.sum((sX.mean_ * self.coef_) / sX.scale_)
            self.coef_ /= sX.scale_

    @staticmethod
    def soft_threshold(X, threshold):
        """Performs the soft-thresholding necessary for coordinate descent
        lasso updates.

        Parameters
        ----------
        X : array-like, shape (n_features, n_samples)
            Matrix to be thresholded.
        threshold : float
            Soft threshold.

        Returns
        -------
        X_soft_threshold : array-like
            Soft thresholded X.
        """
        X_thresholded = (np.abs(X) > threshold).astype('int')
        X_soft_threshold = np.sign(X) * (np.abs(X) - threshold) * X_thresholded
        return X_soft_threshold

    @staticmethod
    def adjusted_response(X, y, coef, intercept=0):
        """Calculates the adjusted response when posing the fitting procedure
        as Iteratively Reweighted Least Squares (Newton update on log
        likelihood).

        Parameters
        ----------
        X : array-like, shape (n_features, n_samples)
            The design matrix.
        y : array-like, shape (n_samples)
            The response vector.
        coef : array-like, shape (n_features)
            Current estimate of the parameters.
        intercept : float
            The current estimate of the intercept.

        Returns
        -------
        w : array-like, shape (n_samples)
            Weights for samples. The log-likelihood for a GLM, when posed as
            a linear regression problem, requires reweighting the samples.
        z : array-like, shape (n_samples)
            Working response. The linearized response when rewriting coordinate
            descent for the Poisson likelihood as a iteratively reweighted
            least squares.
        """
        Xbeta = intercept + np.dot(X, coef)
        w = np.exp(Xbeta)
        z = Xbeta + (y / w) - 1.
        return w, z


class UoI_Poisson(AbstractUoIGeneralizedLinearRegressor, Poisson):
    r"""UoI\ :sub:`Poisson` solver.

    Parameters
    ----------
    n_boots_sel : int
        The number of data bootstraps to use in the selection module.
        Increasing this number will make selection more strict.
    n_boots_est : int
        The number of data bootstraps to use in the estimation module.
        Increasing this number will relax selection and decrease variance.
    selection_frac : float
        The fraction of the dataset to use for training in each resampled
        bootstrap, during the selection module. Small values of this parameter
        imply larger "perturbations" to the dataset.
    estimation_frac : float
        The fraction of the dataset to use for training in each resampled
        bootstrap, during the estimation module. The remaining data is used
        to obtain validation scores. Small values of this parameters imply
        larger "perturbations" to the dataset.
    n_lambdas : int
        The number of regularization values to use for selection.
    alphas : list or ndarray of floats
        The parameter that trades off L1 versus L2 regularization for a given
        lambda.
    stability_selection : int, float, or array-like
        If int, treated as the number of bootstraps that a feature must
        appear in to guarantee placement in selection profile. If float,
        must be between 0 and 1, and is instead the proportion of
        bootstraps. If array-like, must consist of either ints or floats
        between 0 and 1. In this case, each entry in the array-like object
        will act as a separate threshold for placement in the selection
        profile.
    estimation_score : string, "log" | "AIC", | "AICc" | "BIC"
        Objective used to choose the best estimates per bootstrap.
    estimation_target : string, "train" | "test"
        Decide whether to assess the estimation_score on the train or test data
        across each bootstrap. By deafult, a sensible choice is made based on
        the chosen ``estimation_score``.
    solver : string, 'lbfgs' | 'cd'
        The solver to use. Options are 'lbfgs' (orthant-wise LBFGS) and 'cd'
        (coordinate descent).
    warm_start : bool
        When set to ``True``, reuse the solution of the previous call to fit as
        initialization, otherwise, just erase the previous solution
    eps : float
        Length of the lasso path. ``eps=1e-3`` means that
        ``alpha_min / alpha_max = 1e-3``.
    tol : float
        Stopping criteria for solver.
    copy_X : boolean
        If ``True``, X will be copied; else, it may be overwritten.
    fit_intercept : boolean
        Whether to calculate the intercept for this model. If set to ``False``,
        no intercept will be used in calculations (e.g. data is expected to be
        already centered).
    standardize : bool
        If True, the regressors X will be standardized before regression by
        subtracting the mean and dividing by their standard deviations.
    max_iter : int
        Maximum number of iterations for iterative fitting methods.
    random_state : int, RandomState instance, or None
        The seed of the pseudo random number generator that selects a random
        feature to update.  If int, random_state is the seed used by the random
        number generator; If RandomState instance, random_state is the random
        number generator; If None, the random number generator is the
        RandomState instance used by ``np.random``.
    comm : MPI communicator
        If passed, the selection and estimation steps are parallelized.

    Attributes
    ----------
    coef_ : ndarray, shape (n_features,) or (n_targets, n_features)
        Estimated coefficients for the linear regression problem.
    intercept_ : float
        Independent term in the linear model.
    supports_ : ndarray, shape (n_supports, n_features)
        Boolean array indicating whether a given regressor (column) is selected
        for estimation for a given regularization parameter value (row).
    """
    def __init__(self, n_boots_sel=24, n_boots_est=24, n_lambdas=48,
                 alphas=np.array([1.]), selection_frac=0.8,
                 estimation_frac=0.8, stability_selection=1.,
                 estimation_score='log', estimation_target=None,
                 solver='lbfgs', warm_start=True,
                 eps=1e-3, tol=1e-5, copy_X=True, fit_intercept=True,
                 standardize=True, max_iter=1000,
                 random_state=None, comm=None, logger=None):
        super(UoI_Poisson, self).__init__(
            n_boots_sel=n_boots_sel,
            n_boots_est=n_boots_est,
            selection_frac=selection_frac,
            estimation_frac=estimation_frac,
            stability_selection=stability_selection,
            estimation_score=estimation_score,
            estimation_target=estimation_target,
            copy_X=copy_X,
            fit_intercept=fit_intercept,
            random_state=random_state,
            comm=comm,
            logger=logger)
        self.n_lambdas = n_lambdas
        self.alphas = alphas
        self.n_alphas = len(alphas)
        self.warm_start = warm_start
        self.eps = eps
        self.lambdas = None
        self._selection_lm = Poisson(
            fit_intercept=fit_intercept,
            standardize=standardize,
            max_iter=max_iter,
            tol=tol,
            warm_start=warm_start,
            solver=solver)
        # estimation is a Poisson regression with no regularization
        self._estimation_lm = Poisson(
            alpha=0.,
            l1_ratio=1.,
            fit_intercept=fit_intercept,
            max_iter=max_iter,
            tol=tol,
            warm_start=False,
            solver=solver)

    def get_reg_params(self, X, y):
        r"""Calculates the regularization parameters (alpha and lambda) to be
        used for the provided data.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The design matrix.
        y : array-like, shape (n_samples)
            The response vector.

        Returns
        -------
        reg_params : a list of dictionaries
            A list containing dictionaries with the value of each
            (lambda, alpha) describing the type of regularization to impose.
            The keys adhere to scikit-learn's terminology (lambda->alpha,
            alpha->l1_ratio). This allows easy passing into the ElasticNet
            object.
        """
        n_samples = X.shape[0]
        if self.lambdas is None:
            self.lambdas = np.zeros((self.n_alphas, self.n_lambdas))
            # a set of lambdas are generated for each alpha value (l1_ratio in
            # sci-kit learn parlance)
            for alpha_idx, alpha in enumerate(self.alphas):
                # calculate upper bound for lambda sweep
                ybar = y.mean()
                lambda_max = np.max(np.abs(np.dot(X.T, y - ybar)))
                lambda_max /= n_samples * alpha
                self.lambdas[alpha_idx, :] = np.logspace(
                    start=np.log10(lambda_max),
                    stop=np.log10(self.eps * lambda_max),
                    num=self.n_lambdas)

        # place the regularization parameters into a list of dictionaries
        reg_params = list()
        for alpha_idx, alpha in enumerate(self.alphas):
            for lamb_idx, lamb in enumerate(self.lambdas[alpha_idx]):
                # reset the regularization parameter
                reg_params.append(dict(alpha=lamb, l1_ratio=alpha))

        return reg_params

    def _score_predictions(self, metric, fitter, X, y, support, boot_idxs=None):
        """Score, according to some metric, predictions provided by a model.

        The resulting score will be negated if an information criterion is
        specified.

        Parameters
        ----------
        metric : string
            The type of score to run on the prediction. Valid options include
            'r2' (explained variance), 'BIC' (Bayesian information criterion),
            'AIC' (Akaike information criterion), and 'AICc' (corrected AIC).
        fitter : Poisson object
            The Poisson object that has been fit to the data with the
            respective hyperparameters.
        X : ndarray, shape (n_samples, n_features)
            The design matrix.
        y : ndarray, shape (n_samples,)
            The response vector.
        support: ndarray
            The indices of the non-zero features.
        boot_idxs : 2-tuple of array-like objects
            Tuple of (train_idxs, test_idxs) generated from a bootstrap
            sample. If this is specified, then the appropriate set of
            data will be used for evaluating scores: test data for r^2,
            and training data for information criteria

        Returns
        -------
        score : float
            The score.
        """

        # Select the train data
        if boot_idxs is not None:
            X = X[boot_idxs[self._estimation_target]]
            y = y[boot_idxs[self._estimation_target]]

        # for Poisson, use predict_mean to calculate the "predicted" values
        y_pred = fitter.predict_mean(X[:, support])
        # calculate the log-likelihood
        ll = utils.log_likelihood_glm(model='poisson', y_true=y, y_pred=y_pred)
        if metric == 'log':
            score = ll
        # information criteria
        else:
            n_features = np.count_nonzero(support)
            if fitter.intercept_ != 0:
                n_features += 1
            n_samples = X.shape[0]
            if metric == 'BIC':
                score = utils.BIC(n_samples * ll, n_features, n_samples)
            elif metric == 'AIC':
                score = utils.AIC(n_samples * ll, n_features)
            elif metric == 'AICc':
                score = utils.AICc(n_samples * ll, n_features, n_samples)
            else:
                raise ValueError(metric + ' is not a valid metric.')
            # negate the score since lower information criterion is preferable
            score = -score

        return score

    def _pre_fit(self, X, y):
        """Perform standardization, if needed, before fitting."""
        if self.standardize:
            self._X_scaler = StandardScaler()
            X = self._X_scaler.fit_transform(X)
        if y.ndim == 2:
            self.output_dim = y.shape[1]
        else:
            self.output_dim = 1
        return X, y

    def _post_fit(self, X, y):
        """Rescale coefficients, if needed, after fitting."""
        if self.output_dim == 1:
            self.coef_ = np.squeeze(self.coef_)
        if self.standardize:
            sX = self._X_scaler
            if self.fit_intercept:
                self.intercept_ -= np.sum((sX.mean_ * self.coef_) / sX.scale_)
            self.coef_ /= sX.scale_

    def _fit_intercept(self, X, y):
        """"Fit a model with an intercept and fixed coefficients.

        This is used to re-fit the intercept after the coefficients are
        estimated.
        """
        if self.fit_intercept:
            mu = np.exp(np.dot(X, self.coef_.T))
            if np.any(y):
                self.intercept_ = np.log(
                    np.mean(y, keepdims=True) / np.mean(mu)
                )
            else:
                self.intercept_ = np.array(-np.inf)
        else:
            self.intercept_ = np.zeros(1)

    def _fit_intercept_no_features(self, y):
        """"Fit a model with only an intercept.

        This is used in cases where the model has no support selected.
        """
        return PoissonInterceptFitterNoFeatures(y)


class PoissonInterceptFitterNoFeatures(object):
    def __init__(self, y):
        if np.any(y):
            self.intercept_ = np.log(y.mean())
        else:
            self.intercept_ = -np.inf

    def predict(self, X):
        """Predicts the response variable given a design matrix. The output is
        the mode of the Poisson distribution.

        Parameters
        ----------
        X : array_like, shape (n_samples, n_features)
            Design matrix to predict on.

        Returns
        -------
        mode : array_like, shape (n_samples)
            The predicted response values, i.e. the modes.
        """
        mu = np.exp(self.intercept_)
        mode = np.floor(mu)
        return mode

    def predict_mean(self, X):
        """Calculates the mean response variable given a design matrix.

        Parameters
        ----------
        X : array_like, shape (n_samples, n_features)
            Design matrix to predict on.

        Returns
        -------
        mu : array_like, shape (n_samples)
            The predicted response values, i.e. the conditional means.
        """
        mu = np.exp(self.intercept_)
        return mu


def _poisson_loss_and_grad(coef, X, y, l2_penalty, sample_weight=None):
    """Computes the Poisson loss and gradient.

    Parameters
    ----------
    coef : ndarray, shape (n_features,) or (n_features + 1,)
        Coefficient vector.
    X : array-like, shape (n_samples, n_features)
        Design matrix.
    y : ndarray, shape (n_samples,)
        Response vector.
    l2_penalty : float
        Regularization parameter on l2-norm.
    sample_weight : array-like, shape (n_samples,), default None
        Array of weights assigned to the individual samples. If None, then each
        sample is provided an equal weight.

    Returns
    -------
    out : float
        Negative log-likelihood of Poisson GLM.
    grad : ndarray, shape (n_features,) or (n_features + 1,)
        Poisson gradient.
    """
    n_samples, n_features = X.shape
    grad = np.empty_like(coef)

    if sample_weight is None:
        sample_weight = np.ones(n_samples)

    # extract intercept
    if grad.shape[0] > n_features:
        intercept = coef[-1]
        coef = coef[:n_features]
    else:
        intercept = 0

    # calculate likelihood
    eta = intercept + np.dot(X, coef)
    out = -np.sum(sample_weight * (y * eta - np.exp(eta))) / n_samples
    out += 0.5 * l2_penalty * np.dot(coef, coef)

    # gradient of parameters
    y_res = sample_weight * (y - np.exp(eta))
    grad[:n_features] = -np.dot(X.T, y_res) / n_samples + l2_penalty * coef

    # gradient of intercept
    if grad.shape[0] > n_features:
        grad[-1] = -np.mean(y_res)

    return out, grad



"""
Helper functions for loading data and managing arrays across ranks with MPI.
"""
import h5py
import numpy as np

try:
    from mpi4py import MPI
    _np2mpi = {np.dtype(np.float32): MPI.FLOAT,
               np.dtype(np.float64): MPI.DOUBLE,
               np.dtype(np.int): MPI.LONG,
               np.dtype(np.intc): MPI.INT}
except ImportError:
    pass


def check_valid_ndarray(X):
    """Checks whether X is a ndarray and returns a contiguous version.
    """
    if X is None:
        return X
    if not isinstance(X, np.ndarray):
        raise ValueError('Must be a numpy ndarray.')
    return np.ascontiguousarray(X)


def load_data_MPI(h5_name, X_key='X', y_key='y', comm=None, root=0):
    """Load data from an HDF5 file and broadcast it across MPI ranks.

    This is a helper function. It is also possible to load the data
    without this function.

    Parameters
    ----------
    h5_name : str
        Path to h5 file.
    X_key : str
        Key for the features dataset. (default: 'X')
    y_key : str
        Key for the targets dataset. (default: 'y')
    comm : MPI.COMM_WORLD
        MPI communicator.
    root : int
        This rank will load the data from file.

    Returns
    -------
    X : ndarray
        Features on all MPI ranks.
    y : ndarray
        Targets on all MPI ranks.
    """

    if comm is None:
        comm = MPI.COMM_WORLD
    rank = comm.rank
    Xshape = None
    Xdtype = None
    yshape = None
    ydtype = None
    if rank == root:
        with h5py.File(h5_name, 'r') as f:
            X = f[X_key][()]
            Xshape = X.shape
            Xdtype = X.dtype
            y = f[y_key][()]
            yshape = y.shape
            ydtype = y.dtype
    Xshape = comm.bcast(Xshape, root=root)
    Xdtype = comm.bcast(Xdtype, root=root)
    yshape = comm.bcast(yshape, root=root)
    ydtype = comm.bcast(ydtype, root=root)
    if rank != root:
        X = np.empty(Xshape, dtype=Xdtype)
        y = np.empty(yshape, dtype=ydtype)
    comm.Bcast([X, _np2mpi[np.dtype(X.dtype)]], root=root)
    comm.Bcast([y, _np2mpi[np.dtype(y.dtype)]], root=root)
    return X, y


def Bcast_from_root(send, comm=None, root=0):
    """Broadcast an array from root to all MPI ranks.

    Parameters
    ----------
    send : ndarray or None
        Array to send from root to all ranks. send in other ranks
        has no effect.
    comm : MPI.COMM_WORLD
        MPI communicator.
    root : int
        This rank contains the array to send.

    Returns
    -------
    send : ndarray
        Each rank will have a copy of the array from root.
    """

    send = check_valid_ndarray(send)
    if comm is None:
        comm = MPI.COMM_WORLD
    rank = comm.rank
    if rank == 0:
        dtype = send.dtype
        shape = send.shape
    else:
        dtype = None
        shape = None
    shape = comm.bcast(shape, root=root)
    dtype = comm.bcast(dtype, root=root)
    if rank != 0:
        send = np.empty(shape, dtype=dtype)
    comm.Bcast([send, _np2mpi[np.dtype(dtype)]], root=root)
    return send


def Gatherv_rows(send, comm=None, root=0):
    """Concatenate arrays along the first axis using Gatherv on root.

    Parameters
    ----------
    send : ndarray
        The arrays to concatenate. All dimensions must be equal except for the
        first.
    comm : MPI.COMM_WORLD
        MPI communicator.
    root : int
        This rank will contain the Gatherv'ed array.

    Returns
    -------
    rec : ndarray or None
        Gatherv'ed array on root or None on other ranks.
    """

    send = check_valid_ndarray(send)
    if comm is None:
        comm = MPI.COMM_WORLD
    rank = comm.rank
    dtype = send.dtype
    shape = send.shape
    tot = np.zeros(1, dtype=int)

    # Gather the sizes of the first dimension on root
    rank_sizes = comm.gather(shape[0], root=root)
    comm.Reduce(np.array(shape[0], dtype=int),
                [tot, _np2mpi[tot.dtype]], op=MPI.SUM, root=root)
    if rank == root:
        rec_shape = (tot[0],) + shape[1:]
        rec = np.empty(rec_shape, dtype=dtype)
        sizes = [size * np.prod(rec_shape[1:]) for size in rank_sizes]
        disps = np.insert(np.cumsum(sizes), 0, 0)[:-1]
    else:
        rec = None
        sizes = None
        disps = None

    comm.Gatherv(send, [rec, sizes, disps, _np2mpi[dtype]], root=0)
    return rec



import pytest

import numpy as np
from numpy.testing import assert_array_equal
from numpy.testing import assert_raises

from pyuoi.linear_model.utils import stability_selection_to_threshold
from pyuoi.linear_model.utils import intersection

from pyuoi.utils import check_logger

import logging
try:
    from mpi4py import MPI
except ImportError:
    MPI = None


def test_stability_selection_to_threshold_int():
    """Tests whether stability_selection_to_threshold correctly outputs the
    correct threshold when provided a single integer."""

    n_boots_sel = 48
    # stability selection is a single integer
    test_int = 36
    selection_thresholds = stability_selection_to_threshold(
        test_int, n_boots_sel)

    assert_array_equal(selection_thresholds, np.array([36]))


def test_stability_selection_to_threshold_float():
    """Tests whether stability_selection_to_threshold correctly outputs the
    correct threshold when provided a single float."""

    n_boots_sel = 48
    # stability selection is a single float
    test_float = 0.5
    selection_thresholds = stability_selection_to_threshold(
        test_float, n_boots_sel)

    assert_array_equal(selection_thresholds, np.array([24]))


def test_stability_selection_to_threshold_ints():
    """Tests whether stability_selection_to_threshold correctly outputs the
    correct threshold when provided a list of ints."""

    n_boots_sel = 48
    # stability selection is a list of ints
    test_ints = [24, 28, 33, 38, 43, 48]
    selection_thresholds = stability_selection_to_threshold(
        test_ints, n_boots_sel)

    assert_array_equal(
        selection_thresholds,
        np.array([24, 28, 33, 38, 43, 48]))


def test_stability_selection_to_threshold_floats():
    """Tests whether stability_selection_to_threshold correctly outputs the
    correct threshold when provided a list of floats."""
    n_boots_sel = 48
    # stability selection is a list of floats
    test_floats = [0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
    selection_thresholds = stability_selection_to_threshold(
        test_floats, n_boots_sel)

    assert_array_equal(
        selection_thresholds,
        np.array([24, 28, 33, 38, 43, 48]))


def test_stability_selection_to_threshold_ints_np():
    """Tests whether stability_selection_to_threshold correctly outputs the
    correct threshold when provided a numpy array of ints."""

    n_boots_sel = 48
    # stability selection is a list of ints
    test_ints_np = np.array([24, 28, 33, 38, 43, 48])
    selection_thresholds = stability_selection_to_threshold(
        test_ints_np, n_boots_sel)

    assert_array_equal(
        selection_thresholds,
        np.array([24, 28, 33, 38, 43, 48]))


def test_stability_selection_to_threshold_floats_np():
    """Tests whether stability_selection_to_threshold correctly outputs the
    correct threshold when provided a numpy array of ints."""

    n_boots_sel = 48
    # stability selection is a list of floats
    test_floats_np = np.array([0.5, 0.6, 0.7, 0.8, 0.9, 1.0])
    selection_thresholds = stability_selection_to_threshold(
        test_floats_np, n_boots_sel)

    assert_array_equal(
        selection_thresholds,
        np.array([24, 28, 33, 38, 43, 48]))


def test_stability_selection_to_threshold_exceeds_n_bootstraps():
    """Tests whether stability_selection_to_threshold correctly outputs an
    error when provided an input that results in bootstraps exceeding
    n_boots_sel."""

    n_boots_sel = 48
    # stability selection is a list of floats
    test_floats = np.array([0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1])
    test_ints = np.array([24, 28, 33, 38, 43, 48, 52])

    assert_raises(
        ValueError,
        stability_selection_to_threshold,
        test_ints,
        n_boots_sel)

    assert_raises(
        ValueError,
        stability_selection_to_threshold,
        test_floats,
        n_boots_sel)


def test_stability_selection_to_threshold_one_bootstrap():
    """Tests whether stability_selection_to_threshold correctly handles the
    edge case where one bootstrap is requested."""

    n_boots_sel = 1
    # stability selection can only be one value
    threshold = 1

    selection_thresholds = stability_selection_to_threshold(
        n_boots_sel,
        threshold)

    assert_array_equal(
        selection_thresholds,
        np.array([1]))


def test_stability_selection_to_threshold_input_value_error():
    """Tests whether stability_selection_to_threshold properly raises an error
    when it receives objects without ints or floats."""
    n_boots_sel = 48
    stability_selection_list = [0, 1, 'a']
    stability_selection_np_array = np.array([0, 1, 'a'])
    stability_selection_dict = {0: 'a', 1: 'b'}

    assert_raises(
        ValueError,
        stability_selection_to_threshold,
        stability_selection_list,
        n_boots_sel)

    assert_raises(
        ValueError,
        stability_selection_to_threshold,
        stability_selection_np_array,
        n_boots_sel)

    assert_raises(
        ValueError,
        stability_selection_to_threshold,
        stability_selection_dict,
        n_boots_sel)


def test_stability_selection_reject_negative_numbers():
    """Tests whether stability_selection_to_threshold correctly rejects
    negative thresholds."""

    n_boots_sel = 48

    # stability selection is a list of floats
    test_negative = -1 * np.array([24, 28, 33, 38, 43, 48, 52])

    assert_raises(
        ValueError,
        stability_selection_to_threshold,
        test_negative,
        n_boots_sel)


def test_intersection():
    """Tests whether intersection correctly performs a hard intersection."""

    coefs = np.array([
        [[2, 1, -1, 0, 4],
         [4, 0, 2, -1, 5],
         [1, 2, 3, 4, 5]],
        [[2, 0, 0, 0, 0],
         [3, 1, 1, 0, 3],
         [6, 7, 8, 9, 10]],
        [[2, 0, 0, 0, 0],
         [2, -1, 3, 0, 2],
         [2, 4, 6, 8, 9]]])

    true_intersection = np.array([
        [True, False, False, False, False],
        [True, False, True, False, True],
        [True, True, True, True, True]])

    selection_thresholds = np.array([3])
    estimated_intersection = intersection(
        coefs=coefs,
        selection_thresholds=selection_thresholds)

    # we sort the supports since they might not be in the same order
    assert_array_equal(
        np.sort(true_intersection, axis=0),
        np.sort(estimated_intersection, axis=0))


def test_intersection_with_stability_selection_one_threshold():
    """Tests whether intersection correctly performs a soft intersection."""

    coefs = np.array([
        [[2, 1, -1, 0, 4],
         [4, 0, 2, -1, 5],
         [1, 2, 3, 4, 5]],
        [[2, 0, 0, 0, 0],
         [3, 1, 1, 0, 3],
         [6, 7, 8, 9, 10]],
        [[2, 0, 0, 0, 0],
         [2, -1, 3, 0, 2],
         [2, 4, 6, 8, 9]]])

    true_intersection = np.array([
        [True, False, False, False, False],
        [True, True, True, False, True],
        [True, True, True, True, True]])

    selection_thresholds = np.array([2])
    estimated_intersection = intersection(
        coefs=coefs,
        selection_thresholds=selection_thresholds)

    # we sort the supports since they might not be in the same order
    assert_array_equal(
        np.sort(true_intersection, axis=0),
        np.sort(estimated_intersection, axis=0))


def test_intersection_with_stability_selection_multiple_thresholds():
    """Tests whether intersection correctly performs an intersection with
    multiple thresholds. This test also covers the case when there are
    duplicates."""

    coefs = np.array([
        [[2, 1, -1, 0, 4],
         [4, 0, 2, -1, 5],
         [1, 2, 3, 4, 5]],
        [[2, 0, 0, 0, 0],
         [3, 1, 1, 0, 3],
         [6, 7, 8, 9, 10]],
        [[2, 0, 0, 0, 0],
         [2, -1, 3, 0, 2],
         [2, 4, 6, 8, 9]]])

    true_intersection = np.array([
        [True, False, False, False, False],
        [True, True, True, False, True],
        [True, True, True, True, True],
        [True, False, True, False, True]])

    selection_thresholds = np.array([2, 3])
    estimated_intersection = intersection(
        coefs=coefs,
        selection_thresholds=selection_thresholds)

    # we sort the supports since they might not be in the same order
    assert_array_equal(
        np.sort(true_intersection, axis=0),
        np.sort(estimated_intersection, axis=0))


def test_intersection_no_thresholds():
    """Tests that the intersection method correctly calculates the intersection
    using the number of bootstraps as the default selection threshold."""

    coefs = np.array([
        [[2, 1, -1, 0, 4],
         [4, 0, 2, -1, 5],
         [1, 2, 3, 4, 5]],
        [[2, 0, 0, 0, 0],
         [3, 1, 1, 0, 3],
         [6, 7, 8, 9, 10]],
        [[2, 0, 0, 0, 0],
         [2, -1, 3, 0, 2],
         [2, 4, 6, 8, 9]]])

    true_intersection = np.array([
        [True, False, False, False, False],
        [True, True, True, True, True],
        [True, False, True, False, True]])

    estimated_intersection = intersection(
        coefs=coefs,
        selection_thresholds=None)

    # we sort the supports since they might not be in the same order
    assert_array_equal(
        np.sort(true_intersection, axis=0),
        np.sort(estimated_intersection, axis=0))


@pytest.mark.fast
def test_check_logger():
    """Test that check_logger builds logger correctly"""
    ret = check_logger(None, name="test_check_logger")
    assert ret is not None
    assert ret.name == 'test_check_logger'


@pytest.mark.fast
@pytest.mark.skipif(MPI is None, reason='MPI not installed.')
def test_check_logger_mpi():
    """Test that passing in a MPI communicatorj object works with
    check_logger"""
    comm = MPI.COMM_WORLD
    ret = check_logger(None, comm=comm)
    assert ret is not None


@pytest.mark.fast
def test_check_logger_exists():
    """Test that logger returns the argued logger when it gets passed in"""
    logger = logging.getLogger()
    ret = check_logger(logger)
    assert ret is logger



import numpy as np


def stability_selection_to_threshold(stability_selection, n_boots):
    """Converts user inputted stability selection to an array of
    thresholds. These thresholds correspond to the number of bootstraps
    that a feature must appear in to guarantee placement in the selection
    profile.

    Parameters
    ----------
    stability_selection : int, float, or array-like
        If int, treated as the number of bootstraps that a feature must
        appear in to guarantee placement in selection profile. If float,
        must be between 0 and 1, and is instead the proportion of
        bootstraps. If array-like, must consist of either ints or floats
        between 0 and 1. In this case, each entry in the array-like object
        will act as a separate threshold for placement in the selection
        profile.

    n_boots: int
        The number of bootstraps that will be used for selection
    """

    # single float, indicating proportion of bootstraps
    if isinstance(stability_selection, float):
        selection_thresholds = np.array([int(
            stability_selection * n_boots
        )])

    # single int, indicating number of bootstraps
    elif isinstance(stability_selection, int):
        selection_thresholds = np.array([int(
            stability_selection
        )])

    # list, to be converted into numpy array
    elif isinstance(stability_selection, list):
        # list of floats
        if all(isinstance(idx, float) for idx in stability_selection):
            selection_thresholds = \
                n_boots * np.array(stability_selection)

        # list of ints
        elif all(isinstance(idx, int) for idx in stability_selection):
            selection_thresholds = np.array(stability_selection)

        else:
            raise ValueError("Stability selection list must consist of "
                             "floats or ints.")

    # numpy array
    elif isinstance(stability_selection, np.ndarray):
        # np array of floats
        if np.issubdtype(stability_selection.dtype.type, np.floating):
            selection_thresholds = n_boots * stability_selection

        # np array of ints
        elif np.issubdtype(stability_selection.dtype.type, np.integer):
            selection_thresholds = stability_selection

        else:
            raise ValueError("Stability selection array must consist of "
                             "floats or ints.")

    else:
        raise ValueError("Stability selection must be a valid float, int "
                         "or array.")

    # ensure that ensuing list of selection thresholds satisfies
    # the correct bounds
    selection_thresholds = selection_thresholds.astype('int')
    if not (
        np.all(selection_thresholds <= n_boots) and
        np.all(selection_thresholds >= 1)
    ):
        raise ValueError("Stability selection thresholds must be within "
                         "the correct bounds.")

    return selection_thresholds


def intersection(coefs, selection_thresholds=None):
    """Performs the intersection operation on selection coefficients
    using stability selection criteria.

    The coefficients must be provided in the shape
        bootstraps x lambdas x features.
    The intersection operation finds, for each lambda, the features that
    exist in all bootstraps (hard intersection) or in some subset of them
    (the exact subset is provided by selection_thresholds).

    This parameter selection_thresholds provides the number of bootstraps
    that a feature must exist in to pass the intersection. Importantly,
    this function can take intersections with multiple selection_thresholds
    (thus, selection_thresholds is array-like).

    This function then outputs an array of supports, each as a binary mask.
    Only unique supports are provided, so duplicates are tossed out.

    Parameters
    ----------
    coefs : np.ndarray, shape (# bootstraps, # lambdas, # features)
        The coefficients obtained from the selection sweep, corresponding to
        each bootstrap and choice of L1 regularization strength.

    selection_thresholds: array-like, int
        The selection thresholds to perform intersection across. By default,
        use *coefs.shape[0]*.

    Returns
    -------
    supports : np.ndarray, shape (# supports, # features), bool
        A list of supports (each as a binary mask with size n_features)
        obtained by performing the intersection across the coefficients. Each
        support is unique.
    """

    if selection_thresholds is None:
        selection_thresholds = np.array([coefs.shape[0]])

    n_selection_thresholds = len(selection_thresholds)
    n_reg_params = coefs.shape[1]
    n_features = coefs.shape[2]
    supports = np.zeros(
        (n_selection_thresholds, n_reg_params, n_features),
        dtype=bool
    )

    # iterate over each stability selection threshold
    for thresh_idx, threshold in enumerate(selection_thresholds):
        # calculate the support given the specific selection threshold
        supports[thresh_idx, ...] = \
            np.count_nonzero(coefs, axis=0) >= threshold

    # unravel the dimension corresponding to selection thresholds

    supports = np.squeeze(np.reshape(
        supports,
        (n_selection_thresholds * n_reg_params, n_features)
    ))

    supports = np.unique(supports, axis=0)

    return supports



from .linear_model import UoI_Lasso
from .linear_model import UoI_ElasticNet
from .linear_model import UoI_L1Logistic
from .decomposition import UoI_NMF
from .decomposition import UoI_CUR


__all__ = ["UoI_Lasso",
           "UoI_L1Logistic",
           "UoI_ElasticNet",
           "UoI_NMF",
           "UoI_CUR"]

name = "pyuoi"



import numpy as np
from numpy.testing import assert_almost_equal
from numpy.testing import assert_equal
from numpy.testing import assert_raises

from sklearn.datasets import make_regression, make_classification
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.metrics import r2_score, accuracy_score, log_loss

from pyuoi.utils import log_likelihood_glm
from pyuoi.utils import (AIC, BIC, AICc)

from pyuoi.linear_model import (UoI_Lasso, UoI_L1Logistic, UoI_Poisson,
                                UoI_ElasticNet)


def test_ll():
    """Tests that the log-likelihood for generalized linear models is correctly
    calculated."""

    # identity
    y_true = np.array([1, 2, 3])
    y_pred = np.array([np.e + 1, np.e + 2, np.e + 3])
    ll = log_likelihood_glm('normal', y_true, y_pred)
    assert_almost_equal(ll, -4.5)

    # poisson
    y_true = np.array([1 / np.log(2.), 1 / np.log(3.), 1 / np.log(4.)])
    y_pred = np.array([2., 3., 4.])
    ll = log_likelihood_glm('poisson', y_true, y_pred)
    assert_almost_equal(ll, -2)

    # poisson with all zeros
    y_true = np.zeros(3)
    y_pred = np.zeros(3)
    ll = log_likelihood_glm('poisson', y_true, y_pred)
    assert_equal(ll, 0.)

    # poisson with all zeros, but predicted is not all zeros
    y_pred = np.zeros(3)
    y_true = np.array([0., 0., 1.])
    ll = log_likelihood_glm('poisson', y_true, y_pred)
    assert_equal(ll, -np.inf)


def test_ll_error():
    """Tests that the log-likelihood function correctly raises an error when an
    incorrect string is passed as a parameter."""

    y_true = np.array([1., 2., 3.])
    y_pred = np.array([3., 4., 5.])

    assert_raises(ValueError,
                  log_likelihood_glm,
                  'error',
                  y_true,
                  y_pred)


def test_information_criteria():
    """Tests the information criteria (AIC, AICc, BIC) functions."""
    ll = -1.
    n_features = 5
    n_samples = 1000

    aic = AIC(ll, n_features)
    assert_equal(aic, 12.)

    aicc = AICc(ll, n_features, n_samples)
    assert_equal(aicc, 12. + 30. / 497.)

    # additional test: AICc should equal AIC if the number of samples is one
    # greater than the number of features
    aicc = AICc(ll, n_features, n_features + 1)
    assert_equal(aicc, aic)

    bic = BIC(ll, n_features, n_samples)
    assert_equal(bic, 5 * np.log(1000) + 2)


def test_LinearRegressor_scoring_defaults():
    """Tests that the correct default train/test data are being used
    for scoring estimates in UoIAbstractLinearRegressor. Further
    tests that the scoring itself is being done correctly."""
    seed = 5

    X, y = make_regression(n_samples=100, n_features=10, n_informative=10,
                           random_state=seed)

    train_idxs, test_idxs = train_test_split(np.arange(X.shape[0]),
                                             test_size=0.1,
                                             random_state=seed)
    X_train = X[train_idxs]
    y_train = y[train_idxs]

    X_test = X[test_idxs]
    y_test = y[test_idxs]

    fitter = LinearRegression().fit(X_train, y_train)
    support = np.ones(X.shape[1]).astype(bool)
    # r2 - must use test data
    uoi = UoI_Lasso(estimation_score='r2')
    assert(uoi._estimation_target == 1)

    score = uoi._score_predictions('r2', fitter, X, y, support,
                                   (train_idxs, test_idxs))
    assert_equal(r2_score(y_test, fitter.predict(X_test)), score)

    ll = log_likelihood_glm('normal', y_train,
                            fitter.predict(X_train[:, support]))
    # BIC - must use train data
    uoi = UoI_Lasso(estimation_score='BIC')
    assert(uoi._estimation_target == 0)
    score = -1 * uoi._score_predictions('BIC', fitter, X, y, support,
                                        (train_idxs, test_idxs))
    assert_equal(BIC(ll, *X_train.T.shape), score)

    # AIC - must use train data
    uoi = UoI_Lasso(estimation_score='AIC')
    assert(uoi._estimation_target == 0)

    score = -1 * uoi._score_predictions('AIC', fitter, X, y, support,
                                        (train_idxs, test_idxs))
    assert_equal(AIC(ll, X_train.shape[1]), score)

    # AICc - must use train data
    uoi = UoI_Lasso(estimation_score='AICc')
    assert(uoi._estimation_target == 0)

    score = -1 * uoi._score_predictions('AICc', fitter, X, y, support,
                                        (train_idxs, test_idxs))
    assert_equal(AICc(ll, *X_train.T.shape), score)


def test_GeneralizedLinearRegressor_scoring_defaults():
    """Tests that the correct default train/test data are being used
    for scoring estimates in UoIAbstractGeneralizedLinearRegressor. Further
    tests that the scoring itself is being done correctly."""
    seed = 5

    X, y = make_classification(n_samples=100, n_features=3, n_informative=3,
                               n_redundant=0, n_repeated=0, n_classes=3,
                               n_clusters_per_class=2, random_state=seed)

    train_idxs, test_idxs = train_test_split(np.arange(X.shape[0]),
                                             test_size=0.1,
                                             random_state=seed)

    X_train = X[train_idxs]
    y_train = y[train_idxs]

    X_test = X[test_idxs]
    y_test = y[test_idxs]

    fitter = LogisticRegression().fit(X_train, y_train)
    support = np.ones(X.shape[1]).astype(bool)

    # acc - must use test data
    uoi = UoI_L1Logistic(estimation_score='acc')
    assert(uoi._estimation_target == 1)
    uoi.classes_ = np.unique(y)
    score = uoi._score_predictions('acc', fitter, X, y, support,
                                   (train_idxs, test_idxs))
    assert_equal(accuracy_score(y_test, fitter.predict(X_test)), score)

    # log - must use test data. Note the sign difference
    uoi = UoI_L1Logistic(estimation_score='log')
    assert(uoi._estimation_target == 1)
    uoi.classes_ = np.unique(y)
    score = uoi._score_predictions('log', fitter, X, y, support,
                                   (train_idxs, test_idxs))

    y_pred_test = fitter.predict_proba(X_test[:, support])
    assert_equal(log_loss(y_test, y_pred_test, labels=np.unique(y)),
                 -1 * score)

    ll = -log_loss(y_train, fitter.predict_proba(X_train[:, support]),
                   labels=np.unique(y))
    total_ll = ll * X_train.shape[0]
    # BIC - must use train data
    uoi = UoI_L1Logistic(estimation_score='BIC')
    assert(uoi._estimation_target == 0)
    uoi.classes_ = np.unique(y)
    score = -1 * uoi._score_predictions('BIC', fitter, X, y, support,
                                        (train_idxs, test_idxs))
    assert_equal(BIC(total_ll, *X_train.T.shape), score)

    # AIC
    uoi = UoI_L1Logistic(estimation_score='AIC')
    assert(uoi._estimation_target == 0)
    uoi.classes_ = np.unique(y)
    score = -1 * uoi._score_predictions('AIC', fitter, X, y, support,
                                        (train_idxs, test_idxs))
    assert_equal(AIC(total_ll, X_train.shape[1]), score)

    # AICc
    uoi = UoI_L1Logistic(estimation_score='AICc')
    assert(uoi._estimation_target == 0)
    uoi.classes_ = np.unique(y)
    score = -1 * uoi._score_predictions('AICc', fitter, X, y, support,
                                        (train_idxs, test_idxs))
    assert_equal(AICc(total_ll, *X_train.T.shape), score)


def test_estimation_target():
    """Verify the ability for the user to set the estimation taget variable"""

    # Assess r2 on train data
    uoi = UoI_Lasso(estimation_score='r2', estimation_target='train')

    # train gets converted to the index 0
    assert(uoi._estimation_target == 0)

    # Assess BIC on test data
    uoi = UoI_Lasso(estimation_score='BIC', estimation_target='test')

    # Assess r2 on train data
    uoi = UoI_ElasticNet(estimation_score='r2', estimation_target='train')

    # train gets converted to the index 0
    assert(uoi._estimation_target == 0)

    # Assess BIC on test data
    uoi = UoI_ElasticNet(estimation_score='BIC', estimation_target='test')

    assert(uoi._estimation_target == 1)

    uoi = UoI_L1Logistic(estimation_score='acc', estimation_target='train')

    assert(uoi._estimation_target == 0)

    uoi = UoI_L1Logistic(estimation_score='BIC', estimation_target='test')

    assert(uoi._estimation_target == 1)

    uoi = UoI_Poisson(estimation_score='acc', estimation_target='train')

    assert(uoi._estimation_target == 0)

    uoi = UoI_Poisson(estimation_score='BIC', estimation_target='test')

    assert(uoi._estimation_target == 1)



"""
LBFGS and OWL-QN optimization algorithms

Python wrapper around liblbfgs.
"""

cimport numpy as np
import numpy as np
import warnings

np.import_array()   # initialize Numpy

ctypedef enum LineSearchAlgo :
    LBFGS_LINESEARCH_DEFAULT = 0,
    LBFGS_LINESEARCH_MORETHUENTE = 0,
    LBFGS_LINESEARCH_BACKTRACKING_ARMIJO = 1,
    LBFGS_LINESEARCH_BACKTRACKING = 2,
    LBFGS_LINESEARCH_BACKTRACKING_WOLFE = 2,
    LBFGS_LINESEARCH_BACKTRACKING_STRONG_WOLFE = 3,

ctypedef enum ReturnCode:
    LBFGS_ALLZERO = -2048,
    LBFGS_SUCCESS = 0,
    LBFGS_CONVERGENCE = 0,
    LBFGS_STOP,
    LBFGS_ALREADY_MINIMIZED,
    LBFGSERR_UNKNOWNERROR = -1024,
    LBFGSERR_LOGICERROR,
    LBFGSERR_OUTOFMEMORY,
    LBFGSERR_CANCELED,
    LBFGSERR_INVALID_N,
    LBFGSERR_INVALID_N_SSE,
    LBFGSERR_INVALID_X_SSE,
    LBFGSERR_INVALID_EPSILON,
    LBFGSERR_INVALID_TESTPERIOD,
    LBFGSERR_INVALID_DELTA,
    LBFGSERR_INVALID_LINESEARCH,
    LBFGSERR_INVALID_MINSTEP,
    LBFGSERR_INVALID_MAXSTEP,
    LBFGSERR_INVALID_FTOL,
    LBFGSERR_INVALID_WOLFE,
    LBFGSERR_INVALID_GTOL,
    LBFGSERR_INVALID_XTOL,
    LBFGSERR_INVALID_MAXLINESEARCH,
    LBFGSERR_INVALID_ORTHANTWISE,
    LBFGSERR_INVALID_ORTHANTWISE_START,
    LBFGSERR_INVALID_ORTHANTWISE_END,
    LBFGSERR_OUTOFINTERVAL,
    LBFGSERR_INCORRECT_TMINMAX,
    LBFGSERR_ROUNDING_ERROR,
    LBFGSERR_MINIMUMSTEP,
    LBFGSERR_MAXIMUMSTEP,
    LBFGSERR_MAXIMUMLINESEARCH,
    LBFGSERR_MAXIMUMITERATION,
    LBFGSERR_WIDTHTOOSMALL,
    LBFGSERR_INVALIDPARAMETERS,
    LBFGSERR_INCREASEGRADIENT


cdef extern from "lbfgs.h":
    ctypedef double lbfgsfloatval_t
    ctypedef lbfgsfloatval_t* lbfgsconst_p "const lbfgsfloatval_t *"

    ctypedef lbfgsfloatval_t (*lbfgs_evaluate_t)(void *, lbfgsconst_p,
                              lbfgsfloatval_t *, int, lbfgsfloatval_t)
    ctypedef int (*lbfgs_progress_t)(void *, lbfgsconst_p, lbfgsconst_p,
                                     lbfgsfloatval_t, lbfgsfloatval_t,
                                     lbfgsfloatval_t, lbfgsfloatval_t,
                                     int, int, int)

    ctypedef struct lbfgs_parameter_t:
        int m
        lbfgsfloatval_t epsilon
        int past
        lbfgsfloatval_t delta
        int max_iterations
        LineSearchAlgo linesearch
        int max_linesearch
        lbfgsfloatval_t min_step
        lbfgsfloatval_t max_step
        lbfgsfloatval_t ftol
        lbfgsfloatval_t wolfe
        lbfgsfloatval_t gtol
        lbfgsfloatval_t xtol
        lbfgsfloatval_t orthantwise_c
        int orthantwise_start
        int orthantwise_end

    ReturnCode lbfgs(int, lbfgsfloatval_t *, lbfgsfloatval_t *, lbfgs_evaluate_t,
                     lbfgs_progress_t, void *, lbfgs_parameter_t *)

    void lbfgs_parameter_init(lbfgs_parameter_t *)
    lbfgsfloatval_t *lbfgs_malloc(int)
    void lbfgs_free(lbfgsfloatval_t *)


cdef class CallbackData(object):
    cdef object eval_fn
    cdef object progress_fn
    cdef object extra

    def __init__(self, eval_fn, progress_fn, extra):
        self.eval_fn = eval_fn
        self.progress_fn = progress_fn
        self.extra = extra


# Callback into Python evaluation callable.
cdef lbfgsfloatval_t call_eval(void *cb_data_v,
                               lbfgsconst_p x, lbfgsfloatval_t *g,
                               int n, lbfgsfloatval_t step):
    cdef object cb_data
    cdef np.npy_intp tshape[1]

    callback_data = <object>cb_data_v
    (f, progress_fn, shape, args) = callback_data
    tshape[0] = <np.npy_intp>n
    x_array = np.PyArray_SimpleNewFromData(1, tshape, np.NPY_DOUBLE, <void *>x)
    g_array = np.PyArray_SimpleNewFromData(1, tshape, np.NPY_DOUBLE, <void *>g)

    return f(x_array.reshape(shape), g_array.reshape(shape), *args)


# Callback into Python progress reporting callable.
cdef int call_progress(void *cb_data_v,
                       lbfgsconst_p x, lbfgsconst_p g,
                       lbfgsfloatval_t fx,
                       lbfgsfloatval_t xnorm, lbfgsfloatval_t gnorm,
                       lbfgsfloatval_t step, int n, int k, int ls):
    cdef object cb_data
    cdef np.npy_intp tshape[1]

    callback_data = <object>cb_data_v
    (f, progress_fn, shape, args) = callback_data

    if progress_fn:
        tshape[0] = <np.npy_intp>n
        x_array = np.PyArray_SimpleNewFromData(1, tshape, np.NPY_DOUBLE,
                                               <void *>x)
        g_array = np.PyArray_SimpleNewFromData(1, tshape, np.NPY_DOUBLE,
                                               <void *>g)

        r = progress_fn(x_array.reshape(shape), g_array.reshape(shape), fx,
                        xnorm, gnorm, step, k, ls, *args)
        # TODO what happens when the callback returns the wrong type?
        return 0 if r is None else r
    else:
        return 0


# Copy an ndarray to a buffer allocated with lbfgs_malloc; needed to get the
# alignment right for SSE instructions.
cdef lbfgsfloatval_t *aligned_copy(x) except NULL:
    n = x.shape[0]
    x_copy = lbfgs_malloc(n)
    if x_copy is NULL:
        raise MemoryError
    for i in xrange(n):
        x_copy[i] = x[i]
    return x_copy


_LINE_SEARCH_ALGO = {
    'default' : LBFGS_LINESEARCH_DEFAULT,
    'morethuente' : LBFGS_LINESEARCH_MORETHUENTE,
    'armijo' : LBFGS_LINESEARCH_BACKTRACKING_ARMIJO,
    'wolfe' : LBFGS_LINESEARCH_BACKTRACKING_WOLFE,
    'strongwolfe' : LBFGS_LINESEARCH_BACKTRACKING_STRONG_WOLFE
}


_ERROR_MESSAGES = {
    LBFGSERR_UNKNOWNERROR: "Unknown error." ,
    LBFGSERR_LOGICERROR: "Logic error.",
    LBFGSERR_OUTOFMEMORY: "Insufficient memory.",
    LBFGSERR_CANCELED: "The minimization process has been canceled.",
    LBFGSERR_INVALID_N: "Invalid number of variables specified.",
    LBFGSERR_INVALID_N_SSE: "Invalid number of variables (for SSE) specified.",
    LBFGSERR_INVALID_X_SSE: "The array x must be aligned to 16 (for SSE).",
    LBFGSERR_INVALID_EPSILON: "Invalid parameter epsilon specified.",
    LBFGSERR_INVALID_TESTPERIOD: "Invalid parameter past specified.",
    LBFGSERR_INVALID_DELTA: "Invalid parameter delta specified.",
    LBFGSERR_INVALID_LINESEARCH: "Invalid parameter linesearch specified.",
    LBFGSERR_INVALID_MINSTEP: "Invalid parameter max_step specified.",
    LBFGSERR_INVALID_MAXSTEP: "Invalid parameter max_step specified.",
    LBFGSERR_INVALID_FTOL: "Invalid parameter ftol specified.",
    LBFGSERR_INVALID_WOLFE: "Invalid parameter wolfe specified.",
    LBFGSERR_INVALID_GTOL: "Invalid parameter gtol specified.",
    LBFGSERR_INVALID_XTOL: "Invalid parameter xtol specified.",
    LBFGSERR_INVALID_MAXLINESEARCH:
        "Invalid parameter max_linesearch specified.",
    LBFGSERR_INVALID_ORTHANTWISE: "Invalid parameter orthantwise_c specified.",
    LBFGSERR_INVALID_ORTHANTWISE_START:
        "Invalid parameter orthantwise_start specified.",
    LBFGSERR_INVALID_ORTHANTWISE_END:
        "Invalid parameter orthantwise_end specified.",
    LBFGSERR_OUTOFINTERVAL:
        "The line-search step went out of the interval of uncertainty.",
    LBFGSERR_INCORRECT_TMINMAX:
        "A logic error occurred;"
        " alternatively, the interval of uncertainty became too small.",
    LBFGSERR_ROUNDING_ERROR:
        "A rounding error occurred;"
        " alternatively, no line-search step satisfies"
        " the sufficient decrease and curvature conditions.",
    LBFGSERR_MINIMUMSTEP: "The line-search step became smaller than min_step.",
    LBFGSERR_MAXIMUMSTEP: "The line-search step became larger than max_step.",
    LBFGSERR_MAXIMUMLINESEARCH:
        "The line-search routine reaches the maximum number of evaluations.",
    LBFGSERR_MAXIMUMITERATION:
        "The algorithm routine reaches the maximum number of iterations.",
    LBFGSERR_WIDTHTOOSMALL:
        "Relative width of the interval of uncertainty is at most xtol.",
    LBFGSERR_INVALIDPARAMETERS:
        "A logic error (negative line-search step) occurred.",
    LBFGSERR_INCREASEGRADIENT:
        "The current search direction increases the objective function value.",
}


class LBFGSError(Exception):
    pass


class AllZeroLBFGSError(Exception):
    pass


cdef class LBFGS(object):
    """LBFGS algorithm, wrapped in a class to permit setting parameters"""

    cdef lbfgs_parameter_t params

    def __init__(self):
        lbfgs_parameter_init(&self.params)

    LINE_SEARCH_ALGORITHMS = _LINE_SEARCH_ALGO.keys()

    property m:
        def __get__(self) :
            return self.params.m

        def __set__(self, int val):
            self.params.m = val

    property epsilon:
        def __get__(self) :
            return self.params.epsilon

        def __set__(self, double val):
            self.params.epsilon = val

    property past:
        def __get__(self) :
            return self.params.past

        def __set__(self, int val):
            self.params.past = val

    property delta:
        def __get__(self) :
            return self.params.delta

        def __set__(self, double val):
            self.params.delta = val

    property max_iterations:
        def __get__(self) :
            return self.params.max_iterations

        def __set__(self, int val):
            self.params.max_iterations = val

    property linesearch:
        def __get__(self) :
            return self.params.linesearch

        def __set__(self, algorithm):
            self.params.linesearch = _LINE_SEARCH_ALGO[algorithm]

    property max_linesearch:
        def __get__(self) :
            return self.params.max_linesearch

        def __set__(self, int val):
            self.params.max_linesearch = val

    property min_step:
        def __get__(self) :
            return self.params.min_step

        def __set__(self, double val):
            self.params.min_step = val

    property max_step:
        def __get__(self) :
            return self.params.max_step

        def __set__(self, double val):
            self.params.max_step = val

    property ftol:
        def __get__(self) :
            return self.params.ftol

        def __set__(self, double val):
            self.params.ftol = val

    property gtol:
        def __get__(self) :
            return self.params.gtol

        def __set__(self, double val):
            self.params.gtol = val

    property xtol:
        def __get__(self) :
            return self.params.xtol

        def __set__(self, double val):
            self.params.xtol = val

    property wolfe:
        def __get__(self) :
            return self.params.wolfe

        def __set__(self, double val):
            self.params.wolfe = val

    property orthantwise_c:
        def __get__(self) :
            return self.params.orthantwise_c

        def __set__(self, double val):
            self.params.orthantwise_c = val

    property orthantwise_start:
        def __get__(self) :
            return self.params.orthantwise_start

        def __set__(self, int val):
            self.params.orthantwise_start = val

    property orthantwise_end:
        def __get__(self) :
            return self.params.orthantwise_end

        def __set__(self, int val):
            self.params.orthantwise_end = val

    def minimize(self, f, x0, progress=None, args=()):
        """Minimize a function using LBFGS or OWL-QN

        Parameters
        ----------
        f : callable(x, g, *args)
            Computes function to minimize and its gradient.
            Called with the current position x (a numpy.ndarray), a gradient
            vector g (a numpy.ndarray) to be filled in and *args.
            Must return the value at x and set the gradient vector g.
        x0 : array-like
            Initial values. A copy of this array is made prior to optimization.
        progress : callable(x, g, fx, xnorm, gnorm, step, k, num_eval, *args),
                   optional
            If not None, called at each iteration after the call to f with the
            current values of x, g and f(x), the L2 norms of x and g, the line
            search step, the iteration number, the number of evaluations at
            this iteration and args (see below).
            If the return value from this callable is not 0 and not None,
            optimization is stopped and LBFGSError is raised.
        args : sequence
            Arbitrary list of arguments, passed on to f and progress as *args.
        """

        cdef np.npy_intp n
        cdef int n_i
        cdef int r
        cdef lbfgsfloatval_t *x_a
        cdef lbfgsfloatval_t* fx_final = NULL

        if not callable(f):
            raise TypeError("f must be callable, got %s" % type(f))
        if progress is not None and not callable(progress):
            raise TypeError("progress must be callable, got %s" % type(f))

        x0 = np.atleast_1d(x0)
        n = np.product(x0.shape)

        cdef np.npy_intp tshape[1]
        tshape[0] = <np.npy_intp>n

        n_i = n
        if n_i != n:
            raise LBFGSError("Array of %d elements too large to handle" % n)

        x_a = aligned_copy(x0.ravel())

        try:
            callback_data = (f, progress, x0.shape, args)
            r = lbfgs(n, x_a, fx_final, call_eval,
                      call_progress, <void *>callback_data, &self.params)

            if r == LBFGS_SUCCESS or r == LBFGS_ALREADY_MINIMIZED:

                x_array = np.PyArray_SimpleNewFromData(1, tshape, np.NPY_DOUBLE,
                                                       <void *>x_a).copy()

                return x_array.reshape(x0.shape)
            elif r in (LBFGSERR_ROUNDING_ERROR, LBFGSERR_MAXIMUMLINESEARCH) :
                warnings.warn(_ERROR_MESSAGES[r])
                x_array = np.PyArray_SimpleNewFromData(1, tshape, np.NPY_DOUBLE,
                                                       <void *>x_a).copy()

                return x_array.reshape(x0.shape)
            elif r == LBFGSERR_OUTOFMEMORY:
                raise MemoryError
            elif r == LBFGS_ALLZERO:
                raise AllZeroLBFGSError
            else:
                raise LBFGSError(_ERROR_MESSAGES[r])

        finally:
            lbfgs_free(x_a)



import numpy as np

from numpy.testing import (assert_allclose,
                           assert_equal,
                           assert_raises)

from pyuoi.linear_model import (Poisson,
                                UoI_Poisson)
from pyuoi.linear_model.poisson import (_poisson_loss_and_grad,
                                        PoissonInterceptFitterNoFeatures)
from pyuoi.datasets import make_poisson_regression

from sklearn.exceptions import NotFittedError

# poisson GLM model by hand

# design matrix
X = np.array([
    [0.35, 0.84, 0.95, 0.77, 0.88],
    [0.43, 0.76, 0.47, 0.09, 0.34],
    [0.41, 0.40, 0.08, 0.82, 0.49],
    [0.73, 0.93, 0.39, 0.77, 0.72],
    [0.69, 0.88, 0.32, 0.54, 0.26],
    [0.34, 0.10, 0.55, 0.20, 0.20],
    [0.20, 0.15, 0.23, 0.16, 0.74],
    [0.94, 0.08, 0.97, 0.03, 0.48],
    [0.61, 0.55, 0.72, 0.21, 0.27],
    [0.54, 0.21, 0.98, 0.26, 0.01]])

# true parameters
beta = np.array([3., 1., 0., -2., 0.])

# response variable
y = np.array([2, 6, 0, 4, 6, 3, 0, 16, 8, 7])


def test_soft_threshold():
    """Tests the soft threshold function against equivalent output from a
    MATLAB implementation."""

    true_threshold = np.array([
        [0, 0.34, 0.45, 0.27, 0.38],
        [0, 0.26, 0, 0, 0],
        [0, 0, 0, 0.32, 0],
        [0.23, 0.43, 0, 0.27, 0.22],
        [0.19, 0.38, 0, 0.04, 0],
        [0, 0, 0.05, 0, 0],
        [0, 0, 0.0, 0, 0.24],
        [0.44, 0, 0.47, 0, 0],
        [0.11, 0.05, 0.22, 0, 0],
        [0.04, 0, 0.48, 0, 0]])

    assert_allclose(true_threshold,
                    Poisson.soft_threshold(X, 0.5))


def test_adjusted_response():
    """Tests the adjusted response function against equivalent output from a
    MATLAB implementation."""

    w, z = Poisson.adjusted_response(X, y, beta)

    w_true = np.array([
        1.419067548593257, 6.488296399286710, 0.990049833749168,
        4.854955811237434, 6.488296399286709, 2.054433210643888,
        1.537257523548281, 17.115765537145876, 7.099327065156633,
        3.706173712210199])

    z_true = np.array([
        0.759376179437427, 1.794741970890788, -1.01,
        1.403900392819534, 1.794741970890788, 1.180256767879915,
        -0.57, 2.774810655432013, 2.086867367368360,
        2.198740394692808])

    assert_allclose(w_true, w)
    assert_allclose(z_true, z)


def test_predict():
    """Test the predict function in the Poisson class"""
    # design matrix
    X = np.array([[np.log(2.5), -1, -3],
                  [np.log(3.5), -2, -4],
                  [np.log(4.5), -3, -5],
                  [np.log(5.5), -4, -6]])

    poisson = Poisson()

    # test for NotFittedError
    assert_raises(NotFittedError, poisson.predict, X)

    # create "fit"
    poisson.coef_ = np.array([1, 0, 0])
    poisson.intercept_ = 0
    y_pred = poisson.predict(X)
    y_mode = np.array([2, 3, 4, 5])

    # test for predict
    assert_allclose(y_pred, y_mode)


def test_predict_mean():
    """Test the predict function in the Poisson class"""
    # design matrix
    X = np.array([[np.log(2.5), -1, -3],
                  [np.log(3.5), -2, -4],
                  [np.log(4.5), -3, -5],
                  [np.log(5.5), -4, -6]])

    poisson = Poisson()

    # test for NotFittedError
    assert_raises(NotFittedError, poisson.predict_mean, X)

    # create "fit"
    poisson.coef_ = np.array([1, 0, 0])
    poisson.intercept_ = 0
    y_pred = poisson.predict_mean(X)
    y_mean = np.array([2.5, 3.5, 4.5, 5.5])

    # test for predict
    assert_allclose(y_pred, y_mean)


def test_score_predictions():
    """Test the score predictions function in UoI Poisson."""
    X = np.array([[np.log(2), -1, -3],
                  [np.log(3), -2, -4],
                  [np.log(4), -3, -5],
                  [np.log(5), -4, -6]])
    y = 1. / np.log([2., 3., 4., 5.])
    support = np.array([True, False, False])
    n_samples = y.size

    # create fitter by hand
    fitter = Poisson()
    fitter.coef_ = np.array([1])
    fitter.intercept_ = 0

    uoi_fitter = UoI_Poisson()

    # test log-likelihood
    ll = uoi_fitter._score_predictions(
        metric='log',
        fitter=fitter,
        X=X, y=y, support=support)
    assert_allclose(ll, -2.5)

    # test information criteria
    total_ll = ll * n_samples
    aic = uoi_fitter._score_predictions(
        metric='AIC',
        fitter=fitter,
        X=X, y=y, support=support)

    assert_allclose(aic, 2 * total_ll - 2)

    aicc = uoi_fitter._score_predictions(
        metric='AICc',
        fitter=fitter,
        X=X, y=y, support=support)
    assert_allclose(aicc, aic - 2)

    bic = uoi_fitter._score_predictions(
        metric='BIC',
        fitter=fitter,
        X=X, y=y, support=support)
    assert_allclose(bic, 2 * total_ll - np.log(y.size))

    # test invalid metric
    assert_raises(ValueError,
                  uoi_fitter._score_predictions,
                  'fake',
                  fitter,
                  X, y, support)


def test_poisson_intercept_fitter_no_features():
    """Tests the PoissonInterceptFitterNoFeatures class."""
    y = np.array([0, 1, 2])
    poisson = PoissonInterceptFitterNoFeatures(y)

    X = np.random.normal(size=(10, y.size))
    assert_equal(poisson.intercept_, 0.)
    assert_equal(poisson.predict(X), 1.)
    assert_equal(poisson.predict_mean(X), 1.)


def test_poisson_loss_and_grad():
    """Tests the poisson loss and gradient function."""
    n_features = 5
    n_samples = 100

    # test with empty coefficients
    X = np.random.normal(loc=0, scale=1., size=(n_samples, n_features))
    y = np.random.normal(loc=0, scale=1, size=(n_samples))

    loss, grad = _poisson_loss_and_grad(np.zeros(n_features), X, y, 0.)

    assert_equal(loss, 1.)
    assert_allclose(grad, -np.dot(X.T, y - 1) / n_samples)

    # test with non-empty coefficients
    X = np.ones((n_samples, n_features))
    y = np.zeros(n_samples)

    loss, grad = _poisson_loss_and_grad(np.ones(n_features), X, y, 0)
    assert_equal(loss, np.exp(n_features).mean())
    assert_allclose(grad, np.mean(X * np.exp(n_features), axis=0))


def test_poisson_reg_params():
    """Test whether the upper bound on the regularization parameters correctly
    zero out the coefficients."""
    n_features = 5
    n_samples = 1000

    X, y, beta, intercept = make_poisson_regression(
        n_samples=n_samples,
        n_features=n_features,
        n_informative=n_features,
        random_state=2332)

    alpha_sets = [np.array([0.5]), np.array([1.0])]

    for alpha_set in alpha_sets:
        uoi_poisson = UoI_Poisson(alphas=alpha_set)
        reg_params = uoi_poisson.get_reg_params(X, y)
        alpha = reg_params[0]['alpha']
        l1_ratio = reg_params[0]['l1_ratio']

        # check that coefficients get set to zero
        poisson = Poisson(alpha=1.01 * alpha,
                          l1_ratio=l1_ratio,
                          standardize=False,
                          fit_intercept=True)
        poisson.fit(X, y)
        assert_equal(poisson.coef_, 0.)

        # check that coefficients below the bound are not set to zero
        poisson = Poisson(alpha=0.99 * alpha,
                          l1_ratio=l1_ratio,
                          standardize=False,
                          fit_intercept=True)
        poisson.fit(X, y)
        assert np.count_nonzero(poisson.coef_) > 0


def test_poisson_no_intercept():
    """Tests the Poisson fitter with no intercept."""
    n_features = 3
    n_samples = 10000

    # create data
    X, y, beta, _ = make_poisson_regression(n_samples=n_samples,
                                            n_features=n_features,
                                            n_informative=n_features,
                                            beta=np.array([0.5, 1.0, 1.5]),
                                            random_state=2332)

    # lbfgs
    poisson = Poisson(alpha=0., l1_ratio=0., fit_intercept=False,
                      solver='lbfgs', max_iter=5000)
    poisson.fit(X, y)
    assert_allclose(poisson.coef_, beta, rtol=0.5)

    # coordinate descent
    poisson = Poisson(alpha=0., l1_ratio=0., fit_intercept=False, solver='cd',
                      max_iter=5000)
    poisson.fit(X, y)
    assert_allclose(poisson.coef_, beta, rtol=0.5)

    # broken solver
    poisson = Poisson(alpha=0., l1_ratio=0., fit_intercept=False, solver='ruff',
                      max_iter=5000)
    assert_raises(ValueError, poisson.fit, X, y)


def test_poisson_warm_start():
    """Tests if the warm start initializes coefficients correctly."""
    n_features = 3
    n_samples = 10000

    # create data
    X, y, beta, _ = make_poisson_regression(n_samples=n_samples,
                                            n_features=n_features,
                                            n_informative=n_features,
                                            beta=np.array([0.5, 1.0, 1.5]),
                                            random_state=2332)

    # lbfgs
    poisson = Poisson(alpha=0., l1_ratio=0., fit_intercept=False,
                      solver='lbfgs', max_iter=5000, warm_start=True)
    poisson.fit(X, y)
    first_coef = poisson.coef_

    poisson.coef_ = np.zeros(n_features)
    poisson.fit(X, y)
    second_coef = poisson.coef_

    assert_allclose(first_coef, second_coef, rtol=0.1)

    # cd
    poisson = Poisson(alpha=0., l1_ratio=0., fit_intercept=False,
                      solver='cd', max_iter=5000, warm_start=True)
    poisson.fit(X, y)
    first_coef = poisson.coef_

    poisson.coef_ = np.zeros(n_features)
    poisson.fit(X, y)
    second_coef = poisson.coef_

    assert_allclose(first_coef, second_coef, rtol=0.1)


def test_poisson_with_intercept():
    """Tests the Poisson fitter with no intercept."""
    n_features = 3
    n_samples = 10000

    # create data
    X, y, beta, intercept = make_poisson_regression(
        n_samples=n_samples,
        n_features=n_features,
        n_informative=n_features,
        beta=np.array([0.5, 1.0, 1.5]),
        include_intercept=True,
        random_state=2332)

    # lbfgs
    poisson = Poisson(alpha=0., l1_ratio=0., fit_intercept=True,
                      solver='lbfgs', max_iter=5000)
    poisson.fit(X, y)

    assert_allclose(poisson.coef_, beta, rtol=0.5)
    assert_allclose(poisson.intercept_, intercept, rtol=0.5)

    # coordinate descent
    poisson = Poisson(alpha=0., l1_ratio=0., fit_intercept=True, solver='cd',
                      max_iter=5000)
    poisson.fit(X, y)

    assert_allclose(poisson.coef_, beta, rtol=0.5)
    assert_allclose(poisson.intercept_, intercept, rtol=0.5)


def test_poisson_standardize():
    """Tests the Poisson fitter `standardize=True`."""
    n_features = 3
    n_samples = 200

    # create data
    X, y, beta, intercept = make_poisson_regression(
        n_samples=n_samples,
        n_features=n_features,
        n_informative=n_features,
        beta=np.array([0.5, 1.0, 1.5]),
        include_intercept=True,
        random_state=2332)

    # lbfgs
    poisson = Poisson(alpha=0., l1_ratio=0., fit_intercept=True,
                      solver='lbfgs', standardize=True)
    poisson.fit(X, y)


def test_poisson_with_sparsity():
    """Tests the Poisson fitter with no intercept."""
    n_features = 3
    n_samples = 10000

    # create data
    X, y, beta, _ = make_poisson_regression(n_samples=n_samples,
                                            n_features=n_features,
                                            n_informative=n_features,
                                            beta=np.array([0., 1.0, 1.5]),
                                            random_state=2332)

    # lbfgs
    poisson = Poisson(alpha=0.1, l1_ratio=1., fit_intercept=False,
                      solver='lbfgs', max_iter=5000)
    poisson.fit(X, y)

    assert_equal(np.abs(poisson.coef_[0]), 0)

    # coordinate descent
    poisson = Poisson(alpha=0.1, l1_ratio=1., fit_intercept=False, solver='cd',
                      max_iter=5000)
    poisson.fit(X, y)

    assert_equal(np.abs(poisson.coef_[0]), 0.)


def test_poisson_intercept_fitter_all_zeros():
    """Tests that the Poisson intercept fitter properly assigns an intercept
    without warning when the response variable is all zeros."""
    n_samples = 10

    y = np.zeros(n_samples)

    poisson = PoissonInterceptFitterNoFeatures(y)

    assert_equal(poisson.intercept_, -np.inf)


def test_Poisson_response_constant():
    """Test that UoI Poisson correctly fits the data when the response variable
    is constant."""
    n_features = 5
    n_samples = 100

    X = np.random.normal(size=(n_samples, n_features))
    y = np.zeros(n_samples)

    poisson = Poisson()
    poisson.fit(X, y)

    assert_equal(poisson.intercept_, -np.inf)
    assert_equal(poisson.coef_, np.zeros(n_features))

    y += 1.
    poisson.fit(X, y)

    assert_equal(poisson.intercept_, 0.)
    assert_equal(poisson.coef_, np.zeros(n_features))


def test_UoI_Poisson_all_zeros():
    """Test that UoI Poisson correctly fits the data when the response variable
    is all zeros."""
    n_features = 5
    n_samples = 100

    X = np.random.normal(size=(n_samples, n_features))
    y = np.zeros(n_samples)

    poisson = UoI_Poisson()
    poisson.fit(X, y)

    assert_equal(poisson.intercept_, -np.inf)
    assert_equal(poisson.coef_, np.zeros(n_features))


def test_UoI_Poisson():
    """Tests the UoI Poisson fitter with lbfgs solver."""
    n_features = 3
    n_samples = 5000

    # create data
    X, y, beta, _ = make_poisson_regression(n_samples=n_samples,
                                            n_features=n_features,
                                            n_informative=n_features,
                                            beta=np.array([0., 1.0, 1.5]),
                                            random_state=2332)

    # lbfgs
    poisson = UoI_Poisson(n_lambdas=48, n_boots_sel=30, n_boots_est=30,
                          alphas=np.array([1.0]), warm_start=False)
    poisson.fit(X, y)

    assert_allclose(poisson.coef_, beta, atol=0.5)


def test_fit_intercept():
    """Tests whether `include_intercept` in passed through to the linear models.
    """
    poi = UoI_Poisson(fit_intercept=True)
    assert poi._selection_lm.fit_intercept
    assert poi._estimation_lm.fit_intercept

    poi = UoI_Poisson(fit_intercept=False)
    assert not poi._selection_lm.fit_intercept
    assert not poi._estimation_lm.fit_intercept



import numpy as np

from numpy.testing import assert_equal
from numpy.testing import assert_array_equal
from numpy.testing import assert_raises
from pyuoi.decomposition import CUR, UoI_CUR
from pyuoi.decomposition.utils import (column_select,
                                       stability_selection_to_threshold)

X = np.array([
    [0, 0, 0, 4, 2],
    [0, 4, 2, 1, 0],
    [3, 0, 0, 2, 1],
    [2, 2, 0, 1, 0],
    [1, 2, 4, 1, 3],
    [1, 4, 0, 0, 4],
    [3, 3, 4, 0, 0],
    [3, 2, 3, 0, 4],
    [0, 1, 2, 1, 4],
    [1, 4, 0, 2, 4]])


def test_column_select_all():
    """Test that column select function selects all columns when provided the
    entire SVD and a suitable value of c."""
    _, n_features = X.shape
    _, _, V = np.linalg.svd(X)
    column_indices = column_select(V.T, c=5)

    assert_array_equal(column_indices, np.arange(n_features))


def test_column_select():
    """Test that the column select function selects the column with the highest
    leverage score most often."""
    n_samples, n_features = X.shape
    rank = 3
    n_reps = 1000

    _, _, V = np.linalg.svd(X)
    V_subset = V[:rank].T
    column_flags = np.zeros((n_reps, n_features))

    for rep in range(n_reps):
        column_indices = column_select(V_subset, c=1)
        column_flags[rep, column_indices] = 1

    counts = np.sum(column_flags, axis=0)

    assert_equal(np.argmax(counts), np.argmax(np.sum(V_subset**2, axis=1)))


def test_stability_selection_to_threshold_int():
    """Tests whether stability_selection_to_threshold correctly outputs the
    correct threshold when provided a single integer."""

    n_boots_sel = 48
    # stability selection is a single integer
    test_int = 36
    selection_thresholds = stability_selection_to_threshold(test_int,
                                                            n_boots_sel)

    assert_array_equal(selection_thresholds, test_int)


def test_stability_selection_to_threshold_float():
    """Tests whether stability_selection_to_threshold correctly outputs the
    correct threshold when provided a single float."""

    n_boots_sel = 48
    # stability selection is a single float
    test_float = 0.5
    selection_thresholds = stability_selection_to_threshold(test_float,
                                                            n_boots_sel)

    assert_array_equal(selection_thresholds, np.array([24]))


def test_stability_selection_to_threshold_exceeds_n_bootstraps():
    """Tests whether stability_selection_to_threshold correctly outputs an
    error when provided an input that results in bootstraps exceeding
    n_boots_sel."""

    n_boots_sel = 48
    # stability selection is a list of floats
    test_float = 1.1
    test_int = 50

    assert_raises(
        ValueError,
        stability_selection_to_threshold,
        test_int,
        n_boots_sel)

    assert_raises(
        ValueError,
        stability_selection_to_threshold,
        test_float,
        n_boots_sel)


def test_stability_selection_to_threshold_input_value_error():
    """Tests whether stability_selection_to_threshold properly raises an error
    when it receives objects without ints or floats."""
    n_boots_sel = 48
    stability_selection_list = [0, 1, 'a']
    stability_selection_np_array = np.array([0, 1, 'a'])
    stability_selection_dict = {0: 'a', 1: 'b'}

    assert_raises(
        ValueError,
        stability_selection_to_threshold,
        stability_selection_list,
        n_boots_sel)

    assert_raises(
        ValueError,
        stability_selection_to_threshold,
        stability_selection_np_array,
        n_boots_sel)

    assert_raises(
        ValueError,
        stability_selection_to_threshold,
        stability_selection_dict,
        n_boots_sel)


def test_CUR():
    """Tests that the CUR fitter extracts columns correctly."""
    _, n_features = X.shape
    max_k = 3

    cur = CUR(max_k=max_k)

    cur.fit(X, c=3)
    column_indices = cur.column_indices_
    columns = cur.components_

    assert np.setdiff1d(column_indices, np.arange(n_features)).size == 0
    assert_array_equal(X[:, column_indices], columns)


def test_CUR_fit():
    """Tests that the CUR fitter extracts the correct columns."""
    n_features = 5
    n_samples = 30
    max_k = 3

    # matrix has only one non-zero entry
    X = np.zeros((n_samples, n_features))
    X[0, 0] = 1
    true_columns = np.array([0, 2, 3])

    # fit CUR decomposition
    cur = CUR(max_k=max_k)
    X_new = cur.fit_transform(X)

    assert_array_equal(cur.column_indices_, true_columns)
    assert_array_equal(X_new, X[:, true_columns])


def test_UoI_CUR_check_ks_and_cs():
    """Tests the check_ks_and_cs function in UoI_CUR."""
    n_boots = 5
    max_k = 10
    boots_frac = 0.9

    uoi_cur = UoI_CUR(n_boots=n_boots,
                      max_k=max_k,
                      boots_frac=boots_frac)

    # check ks
    ks, cs = uoi_cur.check_ks_and_cs(ks=1)
    assert_array_equal(ks, np.array([1]))
    assert_array_equal(cs, ks + 20)

    ks, cs = uoi_cur.check_ks_and_cs(ks=[1, 2, 3])
    assert_array_equal(ks, np.array([1, 2, 3]))
    assert_array_equal(cs, ks + 20)

    ks, cs = uoi_cur.check_ks_and_cs(ks=None)
    assert_array_equal(ks, 1 + np.arange(max_k))
    assert_array_equal(cs, ks + 20)

    # check cs
    ks, cs = uoi_cur.check_ks_and_cs(ks=[1, 2], cs=[3, 4])
    assert_array_equal(cs, np.array([3, 4]))

    ks, cs = uoi_cur.check_ks_and_cs(ks=[1, 2, 3], cs=1)
    assert_array_equal(cs, np.array([1, 1, 1]))

    ks, cs = uoi_cur.check_ks_and_cs(ks=[1, 2], cs=2.4)
    assert_array_equal(cs, np.array([2.4, 2.4]))

    # value errors for ks
    assert_raises(ValueError, uoi_cur.check_ks_and_cs, -1)
    assert_raises(ValueError, uoi_cur.check_ks_and_cs, [11])
    assert_raises(ValueError, uoi_cur.check_ks_and_cs, [0.1, -1, 2, 12])
    assert_raises(ValueError, uoi_cur.check_ks_and_cs, 2.0)
    assert_raises(ValueError, uoi_cur.check_ks_and_cs, uoi_cur)

    # value errors for cs
    assert_raises(ValueError, uoi_cur.check_ks_and_cs, None, -1)
    assert_raises(ValueError, uoi_cur.check_ks_and_cs, None, [-11])
    assert_raises(ValueError, uoi_cur.check_ks_and_cs, None, np.array([-12]))
    assert_raises(ValueError, uoi_cur.check_ks_and_cs, 1, [2, 3])


def test_UoI_CUR_basic():
    """Test UoI CUR with no bootstrapping."""
    n_samples, n_features = X.shape
    max_k = 3
    n_boots = 1
    boots_frac = 1

    _, _, V = np.linalg.svd(X)
    V_subset = V[:max_k].T

    uoi_cur = UoI_CUR(n_boots=n_boots,
                      max_k=max_k,
                      boots_frac=boots_frac)
    uoi_cur.fit(X, cs=3)

    max_col = np.argmax(np.sum(V_subset**2, axis=1))

    assert (max_col in uoi_cur.column_indices_)


def test_UoI_CUR_fit():
    """Tests that the CUR fitter extracts the correct columns."""
    n_features = 5
    n_samples = 30
    max_k = 3
    n_boots = 10
    boots_frac = 0.95

    # matrix has only one non-zero entry
    X = np.zeros((n_samples, n_features))
    X[0, 0] = 1
    true_columns = np.array([0, 2, 3])

    # fit CUR decomposition
    uoi_cur = UoI_CUR(n_boots=n_boots,
                      max_k=max_k,
                      boots_frac=boots_frac,
                      random_state=2332)
    X_new = uoi_cur.fit_transform(X)

    assert_array_equal(uoi_cur.column_indices_, true_columns)
    assert_array_equal(uoi_cur.components_, X[:, true_columns])
    assert_array_equal(X_new, X[:, true_columns])


def test_UoI_CUR_vs_CUR():
    """Tests that the CUR fitter extracts columns correctly."""
    _, n_features = X.shape
    max_k = 3
    n_boots = 10
    boots_frac = 0.90

    cur = CUR(max_k=max_k,
              random_state=2332)
    cur.fit(X, c=3)

    uoi_cur = UoI_CUR(n_boots=n_boots,
                      max_k=max_k,
                      boots_frac=boots_frac,
                      random_state=2332)
    uoi_cur.fit(X, cs=3, ks=3)

    assert uoi_cur.column_indices_.size <= cur.column_indices_.size



import pytest
import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal_nulp
from sklearn.datasets import make_regression
try:
    from mpi4py import MPI
except ImportError:
    MPI = None

from pyuoi.datasets import make_classification, make_poisson_regression
from pyuoi.linear_model import (UoI_Lasso,
                                UoI_L1Logistic,
                                UoI_ElasticNet,
                                UoI_Poisson)


@pytest.mark.skipif(MPI is None, reason='MPI not installed.')
def test_variable_selection_lasso():
    """Test basic functionality of UoI_Lasso and that it finds right model"""
    X, y, w = make_regression(coef=True, random_state=1)
    lasso = UoI_Lasso(comm=MPI.COMM_WORLD)
    lasso.fit(X, y)
    true_coef = np.nonzero(w)[0]
    fit_coef = np.nonzero(lasso.coef_)[0]
    assert_array_equal(true_coef, fit_coef)
    assert_array_almost_equal_nulp(true_coef, fit_coef)


@pytest.mark.skipif(MPI is None, reason='MPI not installed.')
def test_variable_selection_enet():
    """Test basic functionality of UoI_Lasso and that it finds right model"""
    X, y, w = make_regression(coef=True, random_state=1)
    enet = UoI_ElasticNet(comm=MPI.COMM_WORLD)
    enet.fit(X, y)
    true_coef = np.nonzero(w)[0]
    fit_coef = np.nonzero(enet.coef_)[0]
    assert_array_equal(true_coef, fit_coef)
    assert_array_almost_equal_nulp(true_coef, fit_coef)


@pytest.mark.skipif(MPI is None, reason='MPI not installed.')
def test_l1logistic_binary():
    """Test that binary L1 Logistic runs in the UoI framework."""
    n_inf = 10
    X, y, w, b = make_classification(n_samples=200,
                                     random_state=6,
                                     n_informative=n_inf,
                                     n_features=20,
                                     w_scale=4.,
                                     include_intercept=True)

    l1log = UoI_L1Logistic(random_state=10, comm=MPI.COMM_WORLD).fit(X, y)
    assert (np.sign(abs(w)) == np.sign(abs(l1log.coef_))).mean() >= .7


@pytest.mark.skipif(MPI is None, reason='MPI not installed.')
def test_l1logistic_multiclass():
    """Test that multiclass L1 Logistic runs in the UoI framework when all
       classes share a support."""
    n_features = 20
    n_inf = 10
    X, y, w, b = make_classification(n_samples=200,
                                     random_state=10,
                                     n_classes=5,
                                     n_informative=n_inf,
                                     n_features=n_features,
                                     shared_support=True,
                                     w_scale=4.)
    l1log = UoI_L1Logistic(comm=MPI.COMM_WORLD).fit(X, y)
    assert (np.sign(abs(w)) == np.sign(abs(l1log.coef_))).mean() >= .8


@pytest.mark.skipif(MPI is None, reason='MPI not installed.')
def test_poisson():
    """Test basic functionality of UoI_Lasso and that it finds right model"""
    n_features = 20
    n_inf = 10
    X, y, w, b = make_poisson_regression(n_samples=200,
                                         n_features=n_features,
                                         n_informative=n_inf,
                                         random_state=10)
    poisson = UoI_Poisson(comm=MPI.COMM_WORLD)
    poisson.fit(X, y)
    assert (np.sign(abs(w)) == np.sign(abs(poisson.coef_))).mean() >= .6



"""
LBFGS and OWL-QN optimization algorithms

Python wrapper around liblbfgs.
"""

import warnings
from ._lowlevel import LBFGS, AllZeroLBFGSError  # noqa: F401


def fmin_lbfgs(f, x0, progress=None, args=(), orthantwise_c=0,
               orthantwise_start=0, orthantwise_end=-1, m=10,
               epsilon=1e-5, past=0, delta=0., max_iterations=0,
               line_search="default", max_linesearch=20, min_step=1e-20,
               max_step=1e+20, ftol=1e-4, wolfe=0.9, gtol=0.9, xtol=1e-30):
    """Minimize a function using LBFGS or OWL-QN

     Parameters
    ----------
    f : callable(x, g, *args)
        Computes function to minimize and its gradient.
        Called with the current position x (a numpy.ndarray), a gradient
        vector g (a numpy.ndarray) to be filled in and *args.
        Must return the value at x and set the gradient vector g.

    x0 : array-like
        Initial values. A copy of this array is made prior to optimization.

    progress : callable(x, g, fx, xnorm, gnorm, step, k, num_eval, *args),
               optional
        If not None, called at each iteration after the call to f with the
        current values of x, g and f(x), the L2 norms of x and g, the line
        search step, the iteration number, the number of evaluations at
        this iteration and args (see below).
        If the return value from this callable is not 0 and not None,
        optimization is stopped and LBFGSError is raised.

    args : sequence
        Arbitrary list of arguments, passed on to f and progress as *args.

    orthantwise_c: float, optional (default=0)
        Coefficient for the L1 norm of variables.
        This parameter should be set to zero for standard minimization
        problems. Setting this parameter to a positive value activates
        Orthant-Wise Limited-memory Quasi-Newton (OWL-QN) method, which
        minimizes the objective function F(x) combined with the L1 norm |x|
        of the variables, {F(x) + C |x|}. This parameter is the coefficient
        for the |x|, i.e., C. As the L1 norm |x| is not differentiable at
        zero, the library modifies function and gradient evaluations from
        a client program suitably; a client program thus have only to return
        the function value F(x) and gradients G(x) as usual. The default value
        is zero.

        If orthantwise_c is set, then line_search cannot be the default
        and must be one of 'armijo', 'wolfe', or 'strongwolfe'.

    orthantwise_start: int, optional (default=0)
        Start index for computing L1 norm of the variables.
        This parameter is valid only for OWL-QN method
        (i.e., orthantwise_c != 0). This parameter b (0 <= b < N)
        specifies the index number from which the library computes the
        L1 norm of the variables x,
            |x| := |x_{b}| + |x_{b+1}| + ... + |x_{N}| .
        In other words, variables x_1, ..., x_{b-1} are not used for
        computing the L1 norm. Setting b (0 < b < N), one can protect
        variables, x_1, ..., x_{b-1} (e.g., a bias term of logistic
        regression) from being regularized. The default value is zero.

    orthantwise_end: int, optional (default=-1)
        End index for computing L1 norm of the variables.
        This parameter is valid only for OWL-QN method
        (i.e., orthantwise_c != 0). This parameter e (0 < e <= N)
        specifies the index number at which the library stops computing the
        L1 norm of the variables x,

    m: int, optional, default=6
        The number of corrections to approximate the inverse hessian matrix.
        The L-BFGS routine stores the computation results of previous `m`
        iterations to approximate the inverse hessian matrix of the current
        iteration. This parameter controls the size of the limited memories
        (corrections). The default value is 6. Values less than 3 are
        not recommended. Large values will result in excessive computing time.

    epsilon: float, optional (default=1e-5)
        Epsilon for convergence test.
        This parameter determines the accuracy with which the solution is to
        be found. A minimization terminates when
            ||g|| < \ref epsilon * max(1, ||x||),
        where ||.|| denotes the Euclidean (L2) norm. The default value is
        1e-5.

    past: int, optional (default=0)
        Distance for delta-based convergence test.
        This parameter determines the distance, in iterations, to compute
        the rate of decrease of the objective function. If the value of this
        parameter is zero, the library does not perform the delta-based
        convergence test. The default value is 0.

    delta: float, optional (default=0.)
        Delta for convergence test.
        This parameter determines the minimum rate of decrease of the
        objective function. The library stops iterations when the
        following condition is met:
            (f' - f) / f < delta,
        where f' is the objective value of `past` iterations ago, and f is
        the objective value of the current iteration.
        The default value is 0.

    max_iterations: int, optional (default=0)
        The maximum number of iterations. Setting this parameter to zero
        continues an optimization process until a convergence or error. The
        default value is 0.

    line_search: str, optional (default="default")
        The line search algorithm.
        This parameter specifies a line search algorithm to be used by the
        L-BFGS routine. Possible values are:

        - 'default': same as 'morethuente'
        - 'morethuente': Method proposed by More and Thuente
        - 'armijo': backtracking with Armijo's conditions
        - 'wolfe': backtracking with Wolfe's conditions
        - 'strongwolfe': backtracking with strong Wolfe's conditions

    max_linesearch: int, optional (default=20)
        The maximum number of trials for the line search.
        This parameter controls the number of function and gradients evaluations
        per iteration for the line search routine. The default value is 20.

    min_step: float, optional (default=1e-20)
        The minimum step of the line search routine.
        The default value is 1e-20. This value need not be modified unless
        the exponents are too large for the machine being used, or unless the
        problem is extremely badly scaled (in which case the exponents should
        be increased).

    max_step: float, optional (default=1e20)
        The maximum step of the line search.
        The default value is 1e+20. This value need not be modified unless
        the exponents are too large for the machine being used, or unless the
        problem is extremely badly scaled (in which case the exponents should
        be increased).

    ftol: float, optional (default=1e-4)
        A parameter to control the accuracy of the line search routine.
        The default value is 1e-4. This parameter should be greater
        than zero and smaller than 0.5.

    wolfe: float, optional (default=0.9)
        A coefficient for the Wolfe condition. This parameter is valid only
        when the backtracking line-search algorithm is used with the Wolfe
        condition (`line_search='wolfe'` or `line_search='strongwolfe'`),
        The default value is 0.9. This parameter should be greater
        the `ftol` parameter and smaller than 1.0.

    gtol: float, optional (default=0.9)
        A parameter to control the accuracy of the line search routine.
        The default value is 0.9. If the function and gradient
        evaluations are inexpensive with respect to the cost of the
        iteration (which is sometimes the case when solving very large
        problems) it may be advantageous to set this parameter to a small
        value. A typical small value is 0.1. This parameter should be
        greater than the ftol parameter (1e-4) and smaller than
        1.0.


    xtol: float, optional (default=1e-30)
        The machine precision for floating-point values.
        This parameter must be a positive value set by a client program to
        estimate the machine precision. The line search routine will terminate
        with the status code (::LBFGSERR_ROUNDING_ERROR) if the relative width
        of the interval of uncertainty is less than this parameter.


    """

    # Input validation to make sure defaults with OWL-QN are adapted correctly
    assert orthantwise_c >= 0, "Orthantwise_c cannot be negative"

    if orthantwise_c > 0 and line_search not in ['wolfe', 'default']:
        line_search = 'wolfe'
        warnings.warn("When using OWL-QN, 'wolfe' is the only valid "
                      + "line_search. line_search has been set to 'wolfe'.")
    elif orthantwise_c > 0 and line_search == 'default':
        line_search = 'wolfe'

    opt = LBFGS()
    opt.orthantwise_c = orthantwise_c
    opt.orthantwise_start = orthantwise_start
    opt.orthantwise_end = orthantwise_end
    opt.m = m
    opt.epsilon = epsilon
    opt.past = past
    opt.delta = delta
    opt.max_iterations = max_iterations
    opt.linesearch = line_search
    opt.max_linesearch = max_linesearch
    opt.min_step = min_step
    opt.max_step = max_step
    opt.ftol = ftol
    opt.wolfe = wolfe
    opt.gtol = gtol
    opt.xtol = xtol

    return opt.minimize(f, x0, progress=progress, args=args)
